"""
This module contains I/O operations for loading Sodar files.

"""

import datetime as dt
import re

import fsspec
import numpy as np
import pandas as pd
import xarray as xr

from act.io.noaapsl import filter_list


def read_mfas_sodar(filepath):
    """
    Returns `xarray.Dataset` with stored data and metadata from a user-defined
    Flat Array MFAS Sodar file. More information can be found here:
    https://www.scintec.com/products/flat-array-sodar-mfas/

    Parameters
    ----------
    filepath : str
        Name of file to read.

    Return
    ------
    ds : xarray.Dataset
        Standard Xarray dataset with the data.

    """
    file = fsspec.open(filepath).open()
    lines = file.readlines()
    lines = [x.decode().rstrip()[:] for x in lines]

    # Retrieve number of height values from line 3.
    _, _, len_height = filter_list(lines[3].split()).astype(int)

    # Retrieve metadata
    file_dict, variable_dict = _metadata_retrieval(lines)

    # Retrieve datetimes and time indices from when datetime rows appear
    skip_time_ind = []
    datetimes = []
    fmt = '%Y-%m-%d %H:%M:%S'
    for i, line in enumerate(lines):
        match = re.search(r'\d{4}-\d{2}-\d{2}\ \d{2}:\d{2}:\d{2}', line)
        if match is None:
            continue
        else:
            date_object = dt.datetime.strptime(match.group(0), fmt)
            datetimes.append(date_object)
            skip_time_ind.append(i)

    datetimes = np.delete(datetimes, 0)
    # Create datetime column with matching datetimes to heights
    data_times = pd.DataFrame(datetimes, columns=['Dates'])
    repeat_times = data_times.loc[data_times.index.repeat(len_height)]

    # This is used to pull only actual data.
    # Code can be added as well to read in the metadata from the first few rows.
    skip_meta_ind = np.arange(0, skip_time_ind[1] + 1, 1)
    skip_full_ind = np.append(skip_meta_ind, skip_time_ind)
    skip_full_ind = np.unique(skip_full_ind)

    # Column row appears 1 row after first time, retrieve column names from that.
    columns = np.delete(filter_list(lines[skip_time_ind[1] + 1].split(' ')), 0).tolist()

    # Tmp column allows for the # column to be pushed over and dropped.
    tmp_columns = columns + ['tmp']

    # Parse data to a dataframe skipping rows that aren't data.
    # tmp_columns is used to removed '#' column that causes
    # columns to move over by one.
    df = pd.read_table(
        filepath, sep=r'\s+', skiprows=skip_full_ind, names=tmp_columns, usecols=columns
    )

    df = df[~df['W'].isin(['dir'])].reset_index(drop=True)

    # Set index to datetime column.
    df = df.set_index(repeat_times['Dates'])

    # Convert dataframe to xarray dataset.
    ds = df.to_xarray()

    # Convert height to float.
    ds['z'] = ds.z.astype(float)

    # Convert all variables from string to float.
    ds = ds.astype(float)

    # Convert variables that should be int back to int.
    ds['error'] = ds.error.astype(int)
    ds['PGz'] = ds.PGz.astype(int)

    # Get unique time and height values.
    time_dim = np.unique(ds.Dates.values)
    height_dim = np.unique(ds.z.values)

    # Use unique time and height values to reindex data to be two dimensional.
    ind = pd.MultiIndex.from_product((time_dim, height_dim), names=('time', 'height'))

    # Xarray 2023.9 contains new syntax, adding try and except for
    # previous version.
    try:
        mindex_coords = xr.Coordinates.from_pandas_multiindex(ind, 'Dates')
        ds = ds.assign_coords(mindex_coords).unstack("Dates")
    except AttributeError:
        ds = ds.assign(Dates=ind).unstack("Dates")

    # Add file metadata.
    for key in file_dict.keys():
        ds.attrs[key] = file_dict[key]

    # Add metadata to the attributes of each variable.
    for key in variable_dict.keys():
        ds[key].attrs = variable_dict[key]

    # Change fill values to nans for floats and 0 for ints.
    # We can't use xr.replace as the fill value changes between variables.
    for var in ds.data_vars:
        if var == 'error':
            continue
        elif var == 'PGz':
            data_with_fill = ds[var].values
            data_with_fill[data_with_fill == 99] = 0
            ds[var].values = data_with_fill
        else:
            data_with_fill = ds[var].values
            fill_value = ds[var].attrs['_FillValue']
            data_with_fill[data_with_fill == fill_value] = np.nan
            ds[var].values = data_with_fill

    # Drop z as its already a coordinate and give coordinate the same attributes.
    ds.height.attrs = ds['z'].attrs
    ds = ds.drop_vars('z')

    return ds


def _metadata_retrieval(lines):
    # File format from line 0.
    _format = lines[0]

    # Sodar type from line 2.
    instrument_type = lines[2]

    # Create np.array of lines to use np.argwhere
    line_array = np.array(lines)

    # Retrieve indices of file information and the end of the metadata block.
    file_info_ind = np.argwhere(line_array == '# file information')[0][0]
    file_type_ind = np.argwhere(line_array == '# file type')[0][0]

    # Index the section of file information.
    file_def = line_array[file_info_ind + 2 : file_type_ind - 1]

    # Create a dictionary of file information to be plugged in later to the xarray
    # dataset attributes.
    file_dict = {}
    for line in file_def:
        key, value = filter_list(line.split(':'))
        file_dict[key.strip()] = value.strip()
    file_dict['format'] = _format
    file_dict['instrument_type'] = instrument_type

    # Change values from strings to float where need be.
    file_dict['antenna azimuth angle [deg]'] = float(file_dict['antenna azimuth angle [deg]'])
    file_dict['height above ground [m]'] = float(file_dict['height above ground [m]'])
    file_dict['height above sea level [m]'] = float(file_dict['height above sea level [m]'])

    # Retrieve indices of variable information.
    variable_info_ind = np.argwhere(line_array == '# variable definitions')[0][0]
    data_ind = np.argwhere(line_array == '# beginning of data block')[0][0]

    # Index the section of variable information.
    variable_def = line_array[variable_info_ind + 2 : data_ind - 1]

    # Create a dictionary of variable information to be plugged in later to the xarray
    # variable attributes. Skipping error code as it does not have metadata similar to
    # the rest of the variables.
    variable_dict = {}
    for i, line in enumerate(variable_def):
        if 'error code' in line:
            continue
        else:
            temp_var_dict = {}
            key, symbol, units, _type, error_mask, fill_value = filter_list(line.split('#'))
            temp_var_dict['variable_name'] = key.strip()
            temp_var_dict['symbol'] = symbol.strip()
            temp_var_dict['units'] = units.strip()
            temp_var_dict['type'] = _type.strip()
            temp_var_dict['error_mask'] = error_mask.strip()
            if key.strip() == 'PGz':
                temp_var_dict['_FillValue'] = int(fill_value)
            else:
                temp_var_dict['_FillValue'] = float(fill_value)
            variable_dict[symbol.strip()] = temp_var_dict

    return file_dict, variable_dict



import pytest

import act


def test_noaa_psl():
    result = act.discovery.download_noaa_psl_data(
        site='ctd',
        instrument='Parsivel',
        startdate='20211231',
        enddate='20220101',
        output='./data/',
    )
    assert len(result) == 48

    result = act.discovery.download_noaa_psl_data(
        site='ctd', instrument='Pressure', startdate='20220101', hour='00'
    )
    assert len(result) == 1

    result = act.discovery.download_noaa_psl_data(
        site='ctd', instrument='GpsTrimble', startdate='20220104', hour='00'
    )
    assert len(result) == 6

    types = [
        'Radar S-band Moment',
        'Radar S-band Bright Band',
        '449RWP Bright Band',
        '449RWP Wind',
        '449RWP Sub-Hour Wind',
        '449RWP Sub-Hour Temp',
        '915RWP Wind',
        '915RWP Temp',
        '915RWP Sub-Hour Wind',
        '915RWP Sub-Hour Temp',
    ]
    for t in types:
        result = act.discovery.download_noaa_psl_data(
            site='ctd', instrument=t, startdate='20220601', hour='01'
        )
        assert len(result) == 1

    types = ['Radar FMCW Moment', 'Radar FMCW Bright Band']
    files = [3, 1]
    for i, t in enumerate(types):
        result = act.discovery.download_noaa_psl_data(
            site='bck', instrument=t, startdate='20220101', hour='01'
        )
        assert len(result) == files[i]

    with pytest.raises(ValueError):
        result = act.discovery.download_noaa_psl_data(
            instrument='Parsivel', startdate='20220601', hour='01'
        )
    with pytest.raises(ValueError):
        result = act.discovery.download_noaa_psl_data(
            site='ctd', instrument='dongle', startdate='20220601', hour='01'
        )



"""
Windrose and windbarb timeseries plot
-------------------------------------

This is an example of how to display wind rose and barb timeseries
from multiple days worth of sounding data.

"""

import numpy as np
from arm_test_data import DATASETS
from matplotlib import pyplot as plt

import act

# Read in sonde files
twp_sonde_wildcard_list = [
    'twpsondewnpnC3.b1.20060119.050300.custom.cdf',
    'twpsondewnpnC3.b1.20060119.112000.custom.cdf',
    'twpsondewnpnC3.b1.20060119.163300.custom.cdf',
    'twpsondewnpnC3.b1.20060119.231600.custom.cdf',
    'twpsondewnpnC3.b1.20060120.043800.custom.cdf',
    'twpsondewnpnC3.b1.20060120.111900.custom.cdf',
    'twpsondewnpnC3.b1.20060120.170800.custom.cdf',
    'twpsondewnpnC3.b1.20060120.231500.custom.cdf',
    'twpsondewnpnC3.b1.20060121.051500.custom.cdf',
    'twpsondewnpnC3.b1.20060121.111600.custom.cdf',
    'twpsondewnpnC3.b1.20060121.171600.custom.cdf',
    'twpsondewnpnC3.b1.20060121.231600.custom.cdf',
    'twpsondewnpnC3.b1.20060122.052600.custom.cdf',
    'twpsondewnpnC3.b1.20060122.111500.custom.cdf',
    'twpsondewnpnC3.b1.20060122.171800.custom.cdf',
    'twpsondewnpnC3.b1.20060122.232600.custom.cdf',
    'twpsondewnpnC3.b1.20060123.052500.custom.cdf',
    'twpsondewnpnC3.b1.20060123.111700.custom.cdf',
    'twpsondewnpnC3.b1.20060123.171600.custom.cdf',
    'twpsondewnpnC3.b1.20060123.231500.custom.cdf',
    'twpsondewnpnC3.b1.20060124.051500.custom.cdf',
    'twpsondewnpnC3.b1.20060124.111800.custom.cdf',
    'twpsondewnpnC3.b1.20060124.171700.custom.cdf',
    'twpsondewnpnC3.b1.20060124.231500.custom.cdf',
]
sonde_filenames = [DATASETS.fetch(file) for file in twp_sonde_wildcard_list]
sonde_ds = act.io.arm.read_arm_netcdf(sonde_filenames)

WindDisplay = act.plotting.WindRoseDisplay(sonde_ds, figsize=(8, 10), subplot_shape=(2,))
WindDisplay.plot(
    'deg', 'wspd', spd_bins=np.linspace(0, 25, 5), num_dirs=30, tick_interval=2, subplot_index=(0,)
)

BarbDisplay = act.plotting.TimeSeriesDisplay({'sonde_darwin': sonde_ds}, figsize=(10, 5))
WindDisplay.put_display_in_subplot(BarbDisplay, subplot_index=(1,))
BarbDisplay.plot_time_height_xsection_from_1d_data(
    'rh', 'pres', cmap='coolwarm_r', vmin=0, vmax=100, num_time_periods=25
)

BarbDisplay.plot_barbs_from_spd_dir('wspd', 'deg', 'pres', num_barbs_x=20)
plt.show()



"""
Read and plot a Sodar file
--------------------------

This example shows how to read and display Sodar data from the Argonne
National Laboratory (ANL) ATMOS site.

Author: Zachary Sherman

"""

import matplotlib.pyplot as plt
from arm_test_data import DATASETS

import act

# Call the read_sodar function.
# Example file is a MFAS Sodar at the ATMOS site. More information
# on the sodar can be found here:
# https://www.scintec.com/products/flat-array-sodar-mfas/
filename_sodar = DATASETS.fetch('sodar.20230404.mnd')
ds = act.io.read_mfas_sodar(filename_sodar)

# Create an ACT TimeSeriesDisplay.
display = act.plotting.TimeSeriesDisplay(
    {'Shear, Wind Direction, and Speed at ANL ATMOS': ds}, subplot_shape=(1,), figsize=(15, 5)
)

# Plot shear with a wind barb overlay, while using a color vision
# deficiency (CVD) colormap.
display.plot('shear', subplot_index=(0,), cvd_friendly=True)
display.plot_barbs_from_spd_dir('speed', 'dir')
plt.show()



"""
ACT: The Atmospheric Community Toolkit
======================================

"""

import importlib.metadata as _importlib_metadata

import lazy_loader as lazy

# No more pandas warnings
from pandas.plotting import register_matplotlib_converters

register_matplotlib_converters()

# Import early so these classes are available to the object
from .qc import QCFilter, QCTests, clean  # noqa

# Import the lazy loaded modules
submodules = [
    'corrections',
    'discovery',
    'io',
    'qc',
    'utils',
    'retrievals',
    'plotting',
    'tests',
]
__getattr__, __dir__, _ = lazy.attach(__name__, submodules)

# Get the version
try:
    __version__ = _importlib_metadata.version("act-atmos")
except _importlib_metadata.PackageNotFoundError:
    # package is not installed
    __version__ = "0.0.0"



"""
NOAA FMCW and parsivel plot
---------------------------
ARM and NOAA have campaigns going on in the Crested Butte, CO region
and as part of that campaign NOAA has FMCW radars deployed that could
benefit the broader ARM and NOAA communities. This is an example of
how to plot both a NOAA FMCW PSL and NOAA parsivel two panel plot
observing the same event.

Author: Zach Sherman, Adam Theisen
"""

import matplotlib.pyplot as plt

import act

# Use the ACT downloader to download a file from the
# Kettle Ponds site on 8/01/2022 between 2200 and 2300 UTC.
result_22 = act.discovery.download_noaa_psl_data(
    site='kps', instrument='Radar FMCW Moment', startdate='20220801', hour='22'
)
result_23 = act.discovery.download_noaa_psl_data(
    site='kps', instrument='Radar FMCW Moment', startdate='20220801', hour='23'
)

# Read in the .raw files from both hours. Spectra data are also downloaded.
ds1 = act.io.noaapsl.read_psl_radar_fmcw_moment([result_22[-1], result_23[-1]])

# Read in the parsivel text files.
url = [
    'https://downloads.psl.noaa.gov/psd2/data/realtime/DisdrometerParsivel/Stats/kps/2022/213/kps2221322_stats.txt',
    'https://downloads.psl.noaa.gov/psd2/data/realtime/DisdrometerParsivel/Stats/kps/2022/213/kps2221323_stats.txt',
]
ds2 = act.io.noaapsl.read_psl_parsivel(url)

# Create a TimeSeriesDisplay object using both datasets.
display = act.plotting.TimeSeriesDisplay(
    {'NOAA Site KPS PSL Radar FMCW': ds1, 'NOAA Site KPS Parsivel': ds2},
    subplot_shape=(2,),
    figsize=(10, 10),
)

# Plot PSL Radar followed by the parsivel data.
display.plot(
    'reflectivity_uncalibrated',
    dsname='NOAA Site KPS PSL Radar FMCW',
    cmap='HomeyerRainbow',
    subplot_index=(0,),
)
display.plot(
    'number_density_drops',
    dsname='NOAA Site KPS Parsivel',
    cmap='HomeyerRainbow',
    subplot_index=(1,),
)
# Adjust ylims of parsivel plot.
display.axes[1].set_ylim([0, 10])
plt.show()



"""
Plot a timeseries of sounding data
----------------------------------------------------

This is a simple example for how to plot multiple columns
in a TimeseriesDisplay.

Author: Maxwell Grover
"""

from arm_test_data import DATASETS
from matplotlib import pyplot as plt

import act

# Read in MET files.
met_wildcard_list = [
    'sgpmetE13.b1.20190101.000000.cdf',
    'sgpmetE13.b1.20190102.000000.cdf',
    'sgpmetE13.b1.20190103.000000.cdf',
    'sgpmetE13.b1.20190104.000000.cdf',
    'sgpmetE13.b1.20190105.000000.cdf',
    'sgpmetE13.b1.20190106.000000.cdf',
    'sgpmetE13.b1.20190107.000000.cdf',
]
met_filenames = [DATASETS.fetch(file) for file in met_wildcard_list]
met_ds = act.io.arm.read_arm_netcdf(met_filenames)


# Plot data
display = act.plotting.TimeSeriesDisplay(met_ds)
display.add_subplots((3, 2), figsize=(15, 10))
display.plot('temp_mean', color='tab:red', subplot_index=(0, 0))
display.plot('rh_mean', color='tab:green', subplot_index=(1, 0))
display.plot('wdir_vec_mean', subplot_index=(2, 0))
display.plot('temp_std', color='tab:red', subplot_index=(0, 1))
display.plot('rh_std', color='tab:green', subplot_index=(1, 1))
display.plot('wdir_vec_std', subplot_index=(2, 1))
plt.show()



"""
Example plot using stripes
--------------------------

Plot up climate stripes plots from already
existing climatologies from ARM data.
Author: Adam Theisen

"""

import matplotlib.pyplot as plt

import act

# SGP E13 MET data has already been processed to yearly averages,
# removing data flagged by embedded qc and DQRs
url = 'https://raw.githubusercontent.com/AdamTheisen/ARM-Climatologies/refs/heads/main/results/sgpmetE13.b1_temp_mean_Y.csv'
col_names = ['time', 'temperature', 'count']
ds = act.io.read_csv(url, column_names=col_names, index_col=0, parse_dates=True)

# Drop years with less than 500000 samples
ds = ds.where(ds['count'] > 500000)

# Create plot display
display = act.plotting.TimeSeriesDisplay(ds, figsize=(10, 2))
reference = ['2003-01-01', '2013-01-01']
display.plot_stripes('temperature', reference_period=reference)

plt.show()



import numpy as np

import act


def test_correct_rl():
    # Using ceil data in RL place to save memory
    files = act.tests.sample_files.EXAMPLE_RL1
    ds = act.io.arm.read_arm_netcdf(files)

    ds = act.corrections.raman_lidar.correct_rl(ds, range_normalize_log_values=True)
    np.testing.assert_almost_equal(np.max(ds['depolarization_counts_high'].values), 9.91, decimal=2)
    np.testing.assert_almost_equal(
        np.min(ds['depolarization_counts_high'].values), -7.00, decimal=2
    )
    np.testing.assert_almost_equal(
        np.mean(ds['depolarization_counts_high'].values), -1.45, decimal=2
    )



"""
Read and plot a HYSPLIT trajectory file from a HYSPlIT run.
-----------------------------------------------------------

This example shows how to read and plot a backtrajectory calculated by the NOAA
HYSPLIT model over Houston.

Author: Robert Jackson
"""

import matplotlib.pyplot as plt
from arm_test_data import DATASETS

import act

# Load the data
filename = DATASETS.fetch('houstonaug300.0summer2010080100')
ds = act.io.read_hysplit(filename)

# Use the GeographicPlotDisplay object to make the plot
disp = act.plotting.GeographicPlotDisplay(ds)
disp.geoplot('PRESSURE', cartopy_feature=['STATES', 'OCEAN', 'LAND'])
plt.show()



"""
Modules for reading in NOAA PSL data.
"""

import datetime as dt
import re
from datetime import datetime, timedelta
from itertools import groupby
from os import path as ospath

import fsspec
import numpy as np
import pandas as pd
import xarray as xr
import yaml

from .text import read_csv


def read_psl_wind_profiler(filepath, transpose=True):
    """
    Returns two `xarray.Datasets` with stored data and metadata from a
    user-defined NOAA PSL wind profiler file each containing
    a different mode. This works for both 449 MHz and 915 MHz Weber
    Wuertz and Weber Wuertz sub-hourly files.

    Parameters
    ----------
    filepath : str
        Name of file(s) to read.
    transpose : bool
        True to transpose the data.

    Return
    ------
    mode_one_ds : xarray.Dataset
        Standard Xarray dataset with the first mode data.
    mode_two_ds : xarray.Dataset
        Standard Xarray dataset with the second mode data.

    """
    # Open the file, read in the lines as a list, and return that list
    file = fsspec.open(filepath).open()
    lines = file.readlines()
    lines = [x.decode().rstrip()[:] for x in lines][1:]

    # Separate sections based on the $ separator in the file
    sections_of_file = (list(g) for _, g in groupby(lines, key='$'.__ne__))

    # Count how many lines need to be skipped when reading into pandas
    start_line = 0
    list_of_datasets = []
    for section in sections_of_file:
        if section[0] != '$':
            list_of_datasets.append(
                _parse_psl_wind_lines(filepath, section, line_offset=start_line)
            )
        start_line += len(section)

    # Return two datasets for each mode and the merge of datasets of the
    # same mode.
    mode_one_ds = xr.concat(list_of_datasets[0::2], dim='time')
    mode_two_ds = xr.concat(list_of_datasets[1::2], dim='time')
    if transpose:
        mode_one_ds = mode_one_ds.transpose('HT', 'time')
        mode_two_ds = mode_two_ds.transpose('HT', 'time')
    return mode_one_ds, mode_two_ds


def read_psl_wind_profiler_temperature(filepath, transpose=True):
    """
    Returns `xarray.Dataset` with stored data and metadata from a user-defined
    NOAA PSL wind profiler temperature file.

    Parameters
    ----------
    filepath : str
        Name of file(s) to read.
    transpose : bool
        True to transpose the data.

    Return
    ------
    ds : xarray.Dataset
        Standard Xarray dataset with the data.

    """
    # Open the file, read in the lines as a list, and return that list
    file = fsspec.open(filepath).open()
    lines = file.readlines()
    lines = [x.decode().rstrip()[:] for x in lines][1:]

    # Separate sections based on the $ separator in the file
    sections_of_file = (list(g) for _, g in groupby(lines, key='$'.__ne__))

    # Count how many lines need to be skipped when reading into pandas
    start_line = 0
    list_of_datasets = []
    for section in sections_of_file:
        if section[0] != '$':
            list_of_datasets.append(
                _parse_psl_temperature_lines(filepath, section, line_offset=start_line)
            )
        start_line += len(section)

    # Merge the resultant datasets together
    if transpose:
        return xr.concat(list_of_datasets, dim='time').transpose('HT', 'time')
    else:
        return xr.concat(list_of_datasets, dim='time')


def _parse_psl_wind_lines(filepath, lines, line_offset=0):
    """
    Reads lines related to wind in a psl file.

    Parameters
    ----------
    filepath : str
        Name of file(s) to read.
    lines : list
      List of strings containing the lines to parse.
    line_offset : int (default = 0)
      Offset to start reading the pandas data table.

    Returns
    -------
    ds : xarray.Dataset
      Xarray dataset with wind data.

    """
    # 1 - site
    site = lines[0]

    # 2 - datetype
    datatype, _, version = filter_list(lines[1].split(' '))

    # 3 - station lat, lon, elevation
    latitude, longitude, elevation = filter_list(lines[2].split('  ')).astype(float)

    # 4 - year, month, day, hour, minute, second, utc
    time = parse_date_line(lines[3])

    # 5 - Consensus averaging time, number of beams, number of range gates
    consensus_average_time, number_of_beams, number_of_range_gates = filter_list(
        lines[4].split('  ')
    ).astype(int)

    # 7 - number of coherent integrations, number of spectral averages,
    # pulse width, inner pulse period'
    # Values duplicate as oblique and vertical values
    (
        number_coherent_integrations_obl,
        number_coherent_integrations_vert,
        number_spectral_averages_obl,
        number_spectral_averages_vert,
        pulse_width_obl,
        pulse_width_vert,
        inner_pulse_period_obl,
        inner_pulse_period_vert,
    ) = filter_list(lines[6].split(' ')).astype(int)

    # 8 - full-scale doppler value, delay to first gate, number of gates,
    # spacing of gates. Values duplicate as oblique and vertical values.
    (
        full_scale_doppler_obl,
        full_scale_doppler_vert,
        beam_vertical_correction,
        delay_first_gate_obl,
        delay_first_gate_vert,
        number_of_gates_obl,
        number_of_gates_vert,
        spacing_of_gates_obl,
        spacing_of_gates_vert,
    ) = filter_list(lines[7].split(' ')).astype(float)

    # 9 - beam azimuth (degrees clockwise from north)
    (
        beam_azimuth1,
        beam_elevation1,
        beam_azimuth2,
        beam_elevation2,
        beam_azimuth3,
        beam_elevation3,
    ) = filter_list(lines[8].split(' ')).astype(float)

    beam_azimuth = np.array([beam_azimuth1, beam_azimuth2, beam_azimuth3], dtype='float32')
    beam_elevation = np.array([beam_elevation1, beam_elevation2, beam_elevation3], dtype='float32')

    # Read in the data table section using pandas
    df = pd.read_csv(filepath, skiprows=line_offset + 10, sep=r'\s+')

    # Only read in the number of rows for a given set of gates
    df = df.iloc[: int(number_of_range_gates)]

    # Grab a list of valid columns, except time
    columns = set(list(df.columns)) - {'time'}
    # Set the data types to be floats
    df = df[list(columns)].astype(float)

    # Nan values are encoded as 999999 - let's reflect that
    df = df.replace(999999.0, np.nan)

    # Ensure the height array is stored as a float
    df['HT'] = df.HT.astype(float)

    # Set the height as an index
    df = df.set_index('HT')

    # Rename the count and snr columns more usefully
    df = df.rename(
        columns={
            'RAD': 'RAD1',
            'RAD.1': 'RAD2',
            'RAD.2': 'RAD3',
            'CNT': 'CNT1',
            'CNT.1': 'CNT2',
            'CNT.2': 'CNT3',
            'SNR': 'SNR1',
            'SNR.1': 'SNR2',
            'SNR.2': 'SNR3',
            'QC': 'QC1',
            'QC.1': 'QC2',
            'QC.2': 'QC3',
        }
    )

    # Convert to an xaray dataset
    ds = df.to_xarray()

    # Add attributes to variables
    # Height
    ds['HT'].attrs['long_name'] = 'height_above_ground'
    ds['HT'].attrs['units'] = 'km'

    # Add time to our dataset
    ds['time'] = time

    # Add in our additional attributes
    ds.attrs['site_identifier'] = site.strip()
    ds.attrs['data_type'] = datatype
    ds.attrs['latitude'] = latitude
    ds.attrs['longitude'] = longitude
    ds.attrs['elevation'] = elevation
    ds.attrs['beam_elevation'] = beam_elevation
    ds.attrs['beam_azimuth'] = beam_azimuth
    ds.attrs['revision_number'] = version
    ds.attrs['data_description'] = (
        'https://psl.noaa.gov/data/obs/data/view_data_type_info.php?SiteID=ctd&DataOperationalID=5855&OperationalID=2371'
    )
    ds.attrs['consensus_average_time'] = consensus_average_time
    ds.attrs['oblique-beam_vertical_correction'] = int(beam_vertical_correction)
    ds.attrs['number_of_beams'] = int(number_of_beams)
    ds.attrs['number_of_range_gates'] = int(number_of_range_gates)

    # Handle oblique and vertical attributes.
    ds.attrs['number_of_gates_oblique'] = int(number_of_gates_obl)
    ds.attrs['number_of_gates_vertical'] = int(number_of_gates_vert)
    ds.attrs['number_spectral_averages_oblique'] = int(number_spectral_averages_obl)
    ds.attrs['number_spectral_averages_vertical'] = int(number_spectral_averages_vert)
    ds.attrs['pulse_width_oblique'] = int(pulse_width_obl)
    ds.attrs['pulse_width_vertical'] = int(pulse_width_vert)
    ds.attrs['inner_pulse_period_oblique'] = int(inner_pulse_period_obl)
    ds.attrs['inner_pulse_period_vertical'] = int(inner_pulse_period_vert)
    ds.attrs['full_scale_doppler_value_oblique'] = float(full_scale_doppler_obl)
    ds.attrs['full_scale_doppler_value_vertical'] = float(full_scale_doppler_vert)
    ds.attrs['delay_to_first_gate_oblique'] = int(delay_first_gate_obl)
    ds.attrs['delay_to_first_gate_vertical'] = int(delay_first_gate_vert)
    ds.attrs['spacing_of_gates_oblique'] = int(spacing_of_gates_obl)
    ds.attrs['spacing_of_gates_vertical'] = int(spacing_of_gates_vert)
    return ds


def _parse_psl_temperature_lines(filepath, lines, line_offset=0):
    """
    Reads lines related to temperature in a psl file.

    Parameters
    ----------
    filepath : str
        Name of file(s) to read.
    lines : list
      List of strings containing the lines to parse.
    line_offset : int (default = 0)
      Offset to start reading the pandas data table.

    Returns
    -------
    ds : xarray.Dataset
      Xarray dataset with temperature data.

    """
    # 1 - site
    site = lines[0]

    # 2 - datetype
    datatype, _, version = filter_list(lines[1].split(' '))

    # 3 - station lat, lon, elevation
    latitude, longitude, elevation = filter_list(lines[2].split('  ')).astype(float)

    # 4 - year, month, day, hour, minute, second, utc
    time = parse_date_line(lines[3])

    # 5 - Consensus averaging time, number of beams, number of range gates.
    consensus_average_time, number_of_beams, number_of_range_gates = filter_list(
        lines[4].split('  ')
    ).astype(int)

    # 7 - number of coherent integrations, number of spectral averages,
    # pulse width, inner pulse period.
    (
        number_coherent_integrations,
        number_spectral_averages,
        pulse_width,
        inner_pulse_period,
    ) = filter_list(lines[6].split(' ')).astype(int)

    # 8 - full-scale doppler value, delay to first gate, number of gates,
    # spacing of gates.
    full_scale_doppler, delay_first_gate, number_of_gates, spacing_of_gates = filter_list(
        lines[7].split(' ')
    ).astype(float)

    # 9 - beam azimuth (degrees clockwise from north)
    beam_azimuth, beam_elevation = filter_list(lines[8].split(' ')).astype(float)

    # Read in the data table section using pandas
    df = pd.read_csv(filepath, skiprows=line_offset + 10, sep=r'\s+')

    # Only read in the number of rows for a given set of gates
    df = df.iloc[: int(number_of_gates)]

    # Grab a list of valid columns, except time
    columns = set(list(df.columns)) - {'time'}

    # Set the data types to be floats
    df = df[list(columns)].astype(float)

    # Nan values are encoded as 999999 - let's reflect that
    df = df.replace(999999.0, np.nan)

    # Ensure the height array is stored as a float
    df['HT'] = df.HT.astype(float)

    # Set the height as an index
    df = df.set_index('HT')

    # Rename the count and snr columns more usefully
    df = df.rename(
        columns={
            'CNT': 'CNT_T',
            'CNT.1': 'CNT_Tc',
            'CNT.2': 'CNT_W',
            'SNR': 'SNR_T',
            'SNR.1': 'SNR_Tc',
            'SNR.2': 'SNR_W',
        }
    )

    # Convert to an xaray dataset
    ds = df.to_xarray()

    # Add attributes to variables
    # Height
    ds['HT'].attrs['long_name'] = 'height_above_ground'
    ds['HT'].attrs['units'] = 'km'

    # Temperature
    ds['T'].attrs['long_name'] = 'average_uncorrected_RASS_temperature'
    ds['T'].attrs['units'] = 'degC'
    ds['Tc'].attrs['long_name'] = 'average_corrected_RASS_temperature'
    ds['Tc'].attrs['units'] = 'degC'

    # Vertical motion (w)
    ds['W'].attrs['long_name'] = 'average_vertical_wind'
    ds['W'].attrs['units'] = 'm/s'

    # Add time to our dataset
    ds['time'] = time

    # Add in our additional attributes
    ds.attrs['site_identifier'] = site.strip()
    ds.attrs['data_type'] = datatype
    ds.attrs['latitude'] = latitude
    ds.attrs['longitude'] = longitude
    ds.attrs['elevation'] = elevation
    ds.attrs['beam_elevation'] = beam_elevation
    ds.attrs['beam_azimuth'] = beam_azimuth
    ds.attrs['revision_number'] = version
    ds.attrs['data_description'] = (
        'https://psl.noaa.gov/data/obs/data/view_data_type_info.php?SiteID=ctd&DataOperationalID=5855&OperationalID=2371'
    )
    ds.attrs['consensus_average_time'] = consensus_average_time
    ds.attrs['number_of_beams'] = int(number_of_beams)
    ds.attrs['number_of_gates'] = int(number_of_gates)
    ds.attrs['number_of_range_gates'] = int(number_of_range_gates)
    ds.attrs['number_spectral_averages'] = int(number_spectral_averages)
    ds.attrs['pulse_width'] = pulse_width
    ds.attrs['inner_pulse_period'] = inner_pulse_period
    ds.attrs['full_scale_doppler_value'] = full_scale_doppler
    ds.attrs['spacing_of_gates'] = spacing_of_gates

    return ds


def filter_list(list_of_strings):
    """
    Parses a list of strings, remove empty strings, and return a numpy array
    """
    return np.array(list(filter(None, list_of_strings)))


def parse_date_line(list_of_strings):
    """
    Parses the date line in PSL files
    """
    year, month, day, hour, minute, second, utc_offset = filter_list(
        list_of_strings.split(' ')
    ).astype(int)
    year += 2000
    return datetime(year, month, day, hour, minute, second)


def read_psl_surface_met(filenames, conf_file=None):
    """
    Returns `xarray.Dataset` with stored data and metadata from a user-defined
    NOAA PSL SurfaceMet file.

    Parameters
    ----------
    filenames : str, list of str
        Name of file(s) to read.
    conf_file : str or Pathlib.path
        Default to ./conf/noaapsl_SurfaceMet.yaml
        Filename containing relative or full path to configuration
        YAML file used to describe the file format for each PSL site.
        If the site is not defined in the default file, the default file
        can be copied to a local location, and the missing site added.
        Then point to that updated configuration file. An issue can be
        opened on GitHub to request the missing site to the configuration
        file.

    Return
    ------
    ds :  xarray.Dataset
        Standard Xarray dataset with the data

    """

    if isinstance(filenames, str):
        site = ospath.basename(filenames)[:3]
    else:
        site = ospath.basename(filenames[0])[:3]

    if conf_file is None:
        conf_file = ospath.join(ospath.dirname(__file__), 'conf', 'noaapsl_SurfaceMet.yaml')

    # Read configuration YAML file
    with open(conf_file) as fp:
        try:
            result = yaml.load(fp, Loader=yaml.FullLoader)
        except AttributeError:
            result = yaml.load(fp)

    # Extract dictionary of just corresponding site
    try:
        result = result[site]
    except KeyError:
        raise RuntimeError(
            f"Configuration for site '{site}' currently not available. "
            'You can manually add the site configuration to a copy of '
            'noaapsl_SurfaceMet.yaml and set conf_file= name of copied file '
            'until the site is added.'
        )

    # Extract date and time from filename to use in extracting format from YAML file.
    search_result = re.match(r'[a-z]{3}(\d{2})(\d{3})\.(\d{2})m', ospath.basename(filenames[0]))
    yy, doy, hh = search_result.groups()
    if yy > '70':
        yy = f'19{yy}'
    else:
        yy = f'20{yy}'

    # Extract location information from configuration file.
    try:
        location_info = result['info']
    except KeyError:
        location_info = None

    # Loop through each date range for the site to extract the correct file format from conf file.
    file_datetime = (
        datetime.strptime(f'{yy}-01-01', '%Y-%m-%d')
        + timedelta(int(doy) - 1)
        + timedelta(hours=int(hh))
    )
    for ii in result.keys():
        if ii == 'info':
            continue

        date_range = [
            datetime.strptime(jj, '%Y-%m-%d %H:%M:%S') for jj in result[ii]['_date_range']
        ]
        if file_datetime >= date_range[0] and file_datetime <= date_range[1]:
            result = result[ii]
            del result['_date_range']
            break

    # Read data files by passing in column names from configuration file.
    ds = read_csv(filenames, column_names=list(result.keys()))

    # Calculate numpy datetime64 values from first 4 columns of the data file.
    time = np.array(ds['Year'].values - 1970, dtype='datetime64[Y]')
    day = np.array(np.array(ds['J_day'].values - 1, dtype='timedelta64[D]'))
    hourmin = ds['HoursMinutes'].values + 10000
    hour = [int(str(ii)[1:3]) for ii in hourmin]
    hour = np.array(hour, dtype='timedelta64[h]')
    minute = [int(str(ii)[3:]) for ii in hourmin]
    minute = np.array(minute, dtype='timedelta64[m]')
    time = time + day + hour + minute
    time = time.astype('datetime64[ns]')
    # Update Dataset to use "time" coordinate and assigned calculated times
    ds = ds.assign_coords(index=time)
    ds = ds.rename(index='time')

    # Loop through configuraton dictionary and apply attributes or
    # perform action for specific attributes.
    for var_name in result:
        for key, value in result[var_name].items():
            if key == '_delete' and value is True:
                del ds[var_name]
                continue

            if key == '_type':
                dtype = result[var_name][key]
                ds[var_name] = ds[var_name].astype(dtype)
                continue

            if key == '_missing_value':
                data_values = ds[var_name].values
                data_values[data_values == result[var_name][key]] = np.nan
                ds[var_name].values = data_values
                continue

            ds[var_name].attrs[key] = value

    # Add location information to Dataset
    if location_info is not None:
        ds.attrs['location_description'] = location_info['name']
        for var_name in ['lat', 'lon', 'alt']:
            value = location_info[var_name]['value']
            del location_info[var_name]['value']
            ds[var_name] = xr.DataArray(data=value, attrs=location_info[var_name])

    return ds


def read_psl_parsivel(files):
    """
    Returns `xarray.Dataset` with stored data and metadata from a user-defined
    NOAA PSL parsivel

    Parameters
    ----------
    files : str or list
        Name of file(s) or urls to read.

    Return
    ------
    ds : xarray.Dataset
        Standard Xarray dataset with the data for the parsivel

    """

    # Define the names for the variables
    names = [
        'time',
        'B1',
        'B2',
        'B3',
        'B4',
        'B5',
        'B6',
        'B7',
        'B8',
        'B9',
        'B10',
        'B11',
        'B12',
        'B13',
        'B14',
        'B15',
        'B16',
        'B17',
        'B18',
        'B19',
        'B20',
        'B21',
        'B22',
        'B23',
        'B24',
        'B25',
        'B26',
        'B27',
        'B28',
        'B29',
        'B30',
        'B31',
        'B32',
        'blackout',
        'good',
        'bad',
        'number_detected_particles',
        'precip_rate',
        'precip_amount',
        'precip_accumulation',
        'equivalent_radar_reflectivity',
        'number_in_error',
        'dirty',
        'very_dirty',
        'damaged',
        'laserband_amplitude',
        'laserband_amplitude_stdev',
        'sensor_temperature',
        'sensor_temperature_stdev',
        'sensor_voltage',
        'sensor_voltage_stdev',
        'heating_current',
        'heating_current_stdev',
        'number_rain_particles',
        'number_non_rain_particles',
        'number_ambiguous_particles',
        'precip_type',
    ]

    # Define the particle sizes and class width sizes based on
    # https://psl.noaa.gov/data/obs/data/view_data_type_info.php?SiteID=ctd&DataOperationalID=5890
    vol_equiv_diam = [
        0.062,
        0.187,
        0.312,
        0.437,
        0.562,
        0.687,
        0.812,
        0.937,
        1.062,
        1.187,
        1.375,
        1.625,
        1.875,
        2.125,
        2.375,
        2.75,
        3.25,
        3.75,
        4.25,
        4.75,
        5.5,
        6.5,
        7.5,
        8.5,
        9.5,
        11.0,
        13.0,
        15.0,
        17.0,
        19.0,
        21.5,
        24.5,
    ]
    class_size_width = [
        0.125,
        0.125,
        0.125,
        0.125,
        0.125,
        0.125,
        0.125,
        0.125,
        0.125,
        0.125,
        0.250,
        0.250,
        0.250,
        0.250,
        0.250,
        0.5,
        0.5,
        0.5,
        0.5,
        0.5,
        1.0,
        1.0,
        1.0,
        1.0,
        1.0,
        2.0,
        2.0,
        2.0,
        2.0,
        2.0,
        3.0,
        3.0,
    ]

    if not isinstance(files, list):
        files = [files]

    # Loop through each file or url and append the dataframe into data for concatenations
    data = []
    end_time = []
    for f in files:
        df = pd.read_table(f, skiprows=[0, 1, 2], names=names, index_col=0, sep=r'\s+')
        # Reading the table twice to get the date so it can be parsed appropriately
        date = pd.read_table(f, nrows=0).to_string().split(' ')[-3]
        time = df.index
        start_time = []
        form = '%y%j%H:%M:%S:%f'
        for t in time:
            start_time.append(pd.to_datetime(date + ':' + t.split('-')[0], format=form))
            end_time.append(pd.to_datetime(date + ':' + t.split('-')[1], format=form))
        df.index = start_time
        data.append(df)

    df = pd.concat(data)

    # Create a 2D size distribution variable from all the B* variables
    dsd = []
    for n in names:
        if 'B' not in n:
            continue
        dsd.append(list(df[n]))

    # Convert the dataframe to xarray DataSet and add variables
    ds = df.to_xarray()
    ds = ds.rename({'index': 'time'})
    long_name = 'Drop Size Distribution'
    attrs = {'long_name': long_name, 'units': 'count'}
    da = xr.DataArray(
        np.transpose(dsd),
        dims=['time', 'particle_size'],
        coords=[ds['time'].values, vol_equiv_diam],
    )
    ds['number_density_drops'] = da

    attrs = {'long_name': 'Particle class size average', 'units': 'mm'}
    da = xr.DataArray(
        class_size_width, dims=['particle_size'], coords=[vol_equiv_diam], attrs=attrs
    )
    ds['class_size_width'] = da

    attrs = {'long_name': 'Class size width', 'units': 'mm'}
    da = xr.DataArray(vol_equiv_diam, dims=['particle_size'], coords=[vol_equiv_diam], attrs=attrs)
    ds['particle_size'] = da

    attrs = {'long_name': 'End time of averaging interval'}
    da = xr.DataArray(end_time, dims=['time'], coords=[ds['time'].values], attrs=attrs)
    ds['interval_end_time'] = da

    # Define the attribuets and metadata and add into the DataSet
    attrs = {
        'blackout': {
            'long_name': 'Number of samples excluded during PC clock sync',
            'units': 'count',
        },
        'good': {'long_name': 'Number of samples that passed QC checks', 'units': 'count'},
        'bad': {'long_name': 'Number of samples that failed QC checks', 'units': 'count'},
        'number_detected_particles': {
            'long_name': 'Total number of detected particles',
            'units': 'count',
        },
        'precip_rate': {'long_name': 'Precipitation rate', 'units': 'mm/hr'},
        'precip_amount': {'long_name': 'Interval accumulation', 'units': 'mm'},
        'precip_accumulation': {'long_name': 'Event accumulation', 'units': 'mm'},
        'equivalent_radar_reflectivity': {'long_name': 'Radar Reflectivity', 'units': 'dB'},
        'number_in_error': {
            'long_name': 'Number of samples that were reported dirt, very dirty, or damaged',
            'units': 'count',
        },
        'dirty': {
            'long_name': 'Laser glass is dirty but measurement is still possible',
            'units': 'unitless',
        },
        'very_dirty': {
            'long_name': 'Laser glass is dirty, partially covered no further measurements are possible',
            'units': 'unitless',
        },
        'damaged': {'long_name': 'Laser damaged', 'units': 'unitless'},
        'laserband_amplitude': {
            'long_name': 'Average signal amplitude of the laser strip',
            'units': 'unitless',
        },
        'laserband_amplitude_stdev': {
            'long_name': 'Standard deviation of the signal amplitude of the laser strip',
            'units': 'unitless',
        },
        'sensor_temperature': {'long_name': 'Average sensor temperature', 'units': 'degC'},
        'sensor_temperature_stdev': {
            'long_name': 'Standard deviation of sensor temperature',
            'units': 'degC',
        },
        'sensor_voltage': {'long_name': 'Sensor power supply voltage', 'units': 'V'},
        'sensor_voltage_stdev': {
            'long_name': 'Standard deviation of the sensor power supply voltage',
            'units': 'V',
        },
        'heating_current': {'long_name': 'Average heating system current', 'units': 'A'},
        'heating_current_stdev': {
            'long_name': 'Standard deviation of heating system current',
            'units': 'A',
        },
        'number_rain_particles': {
            'long_name': 'Number of particles detected as rain',
            'units': 'unitless',
        },
        'number_non_rain_particles': {
            'long_name': 'Number of particles detected not as rain',
            'units': 'unitless',
        },
        'number_ambiguous_particles': {
            'long_name': 'Number of particles detected as ambiguous',
            'units': 'unitless',
        },
        'precip_type': {
            'long_name': 'Precipitation type (1=rain; 2=mixed; 3=snow)',
            'units': 'unitless',
        },
        'number_density_drops': {'long_name': 'Drop Size Distribution', 'units': 'count'},
    }

    for v in ds:
        if v in attrs:
            ds[v].attrs = attrs[v]

    return ds


def read_psl_radar_fmcw_moment(files):
    """
    Returns `xarray.Dataset` with stored data and metadata from
    NOAA PSL FMCW Radar files. See References section for details.

    Parameters
    ----------
    files : str or list
        Name of file(s) to read.  Currently does not support reading URLs but files can
        be downloaded easily using the act.discovery.download_noaa_psl_data function.

    Return
    ------
    ds : xarray.Dataset
        Standard Xarray dataset with the data for the parsivel

    References
    ----------
    Johnston, Paul E., James R. Jordan, Allen B. White, David A. Carter, David M. Costa, and Thomas E. Ayers.
        "The NOAA FM-CW snow-level radar." Journal of Atmospheric and Oceanic Technology 34, no. 2 (2017): 249-267.

    """

    ds = _parse_psl_radar_moments(files)

    return ds


def read_psl_radar_sband_moment(files):
    """
    Returns `xarray.Dataset` with stored data and metadata from
    NOAA PSL S-band Radar files.

    Parameters
    ----------
    files : str or list
        Name of file(s) to read.  Currently does not support reading URLs but files can
        be downloaded easily using the act.discovery.download_noaa_psl_data function.

    Return
    ------
    ds : xarray.Dataset
        Standard Xarray dataset with the data for the parsivel

    """

    ds = _parse_psl_radar_moments(files)

    return ds


def _parse_psl_radar_moments(files):
    """
    Returns `xarray.Dataset` with stored data and metadata from
    NOAA PSL FMCW and S-Band Radar files.

    Parameters
    ----------
    files : str or list
        Name of file(s) to read.  Currently does not support reading URLs but files can
        be downloaded easily using the act.discovery.download_noaa_psl_data function.

    Return
    ------
    ds : xarray.Dataset
        Standard Xarray dataset with the data for the parsivel

    """
    # Set the initial dictionary to convert to xarray dataset
    data = {
        'site': {'dims': ['file'], 'data': [], 'attrs': {'long_name': 'NOAA site code'}},
        'lat': {
            'dims': ['file'],
            'data': [],
            'attrs': {'long_name': 'North Latitude', 'units': 'degree_N'},
        },
        'lon': {
            'dims': ['file'],
            'data': [],
            'attrs': {'long_name': 'East Longitude', 'units': 'degree_E'},
        },
        'alt': {
            'dims': ['file'],
            'data': [],
            'attrs': {'long_name': 'Altitude above mean sea level', 'units': 'm'},
        },
        'freq': {
            'dims': ['file'],
            'data': [],
            'attrs': {'long_name': 'Operating Frequency; Ignore for FMCW', 'units': 'Hz'},
        },
        'azimuth': {
            'dims': ['time'],
            'data': [],
            'attrs': {'long_name': 'Azimuth angle', 'units': 'deg'},
        },
        'elevation': {
            'dims': ['time'],
            'data': [],
            'attrs': {'long_name': 'Elevation angle', 'units': 'deg'},
        },
        'beam_direction_code': {
            'dims': ['time'],
            'data': [],
            'attrs': {'long_name': 'Beam direction code', 'units': ''},
        },
        'year': {'dims': ['time'], 'data': [], 'attrs': {'long_name': '2-digit year', 'units': ''}},
        'day_of_year': {
            'dims': ['time'],
            'data': [],
            'attrs': {'long_name': 'Day of the year', 'units': ''},
        },
        'hour': {
            'dims': ['time'],
            'data': [],
            'attrs': {'long_name': 'Hour of the day', 'units': ''},
        },
        'minute': {'dims': ['time'], 'data': [], 'attrs': {'long_name': 'Minutes', 'units': ''}},
        'second': {'dims': ['time'], 'data': [], 'attrs': {'long_name': 'Seconds', 'units': ''}},
        'interpulse_period': {
            'dims': ['time'],
            'data': [],
            'attrs': {'long_name': 'Interpulse Period', 'units': 'ms'},
        },
        'pulse_width': {
            'dims': ['time'],
            'data': [],
            'attrs': {'long_name': 'Pulse width', 'units': 'ns'},
        },
        'first_range_gate': {
            'dims': ['time'],
            'data': [],
            'attrs': {'long_name': 'Range to first range gate', 'units': 'm'},
        },
        'range_between_gates': {
            'dims': ['time'],
            'data': [],
            'attrs': {'long_name': 'Distance between range gates', 'units': 'm'},
        },
        'n_gates': {
            'dims': ['time'],
            'data': [],
            'attrs': {'long_name': 'Number of range gates', 'units': 'count'},
        },
        'n_coherent_integration': {
            'dims': ['time'],
            'data': [],
            'attrs': {'long_name': 'Number of cohrent integration', 'units': 'count'},
        },
        'n_averaged_spectra': {
            'dims': ['time'],
            'data': [],
            'attrs': {'long_name': 'Number of average spectra', 'units': 'count'},
        },
        'n_points_spectrum': {
            'dims': ['time'],
            'data': [],
            'attrs': {'long_name': 'Number of points in spectra', 'units': 'count'},
        },
        'n_code_bits': {
            'dims': ['time'],
            'data': [],
            'attrs': {'long_name': 'Number of code bits', 'units': 'count'},
        },
        'radial_velocity': {
            'dims': ['time', 'range'],
            'data': [],
            'attrs': {'long_name': 'Radial velocity', 'units': 'm/s'},
        },
        'snr': {
            'dims': ['time', 'range'],
            'data': [],
            'attrs': {'long_name': 'Signal-to-noise ratio - not range corrected', 'units': 'dB'},
        },
        'signal_power': {
            'dims': ['time', 'range'],
            'data': [],
            'attrs': {'long_name': 'Signal Power - not range corrected', 'units': 'dB'},
        },
        'spectral_width': {
            'dims': ['time', 'range'],
            'data': [],
            'attrs': {'long_name': 'Spectral width', 'units': 'm/s'},
        },
        'noise_amplitude': {
            'dims': ['time', 'range'],
            'data': [],
            'attrs': {'long_name': 'noise_amplitude', 'units': 'dB'},
        },
        'qc_variable': {
            'dims': ['time', 'range'],
            'data': [],
            'attrs': {'long_name': 'QC Value - not used', 'units': ''},
        },
        'time': {'dims': ['time'], 'data': [], 'attrs': {'long_name': 'Datetime', 'units': ''}},
        'range': {'dims': ['range'], 'data': [], 'attrs': {'long_name': 'Range', 'units': 'm'}},
        'reflectivity_uncalibrated': {
            'dims': ['time', 'range'],
            'data': [],
            'attrs': {'long_name': 'Range', 'units': 'dB'},
        },
    }

    # Separate out the names as they will be accessed in different parts of the code
    h1_names = ['site', 'lat', 'lon', 'alt', 'freq']
    h2_names = [
        'azimuth',
        'elevation',
        'beam_direction_code',
        'year',
        'day_of_year',
        'hour',
        'minute',
        'second',
    ]
    h3_names = [
        'interpulse_period',
        'pulse_width',
        'first_range_gate',
        'range_between_gates',
        'n_gates',
        'n_coherent_integration',
        'n_averaged_spectra',
        'n_points_spectrum',
        'n_code_bits',
    ]
    names = {
        'radial_velocity': 0,
        'snr': 1,
        'signal_power': 2,
        'spectral_width': 3,
        'noise_amplitude': 4,
        'qc_variable': 5,
    }

    # If file is a string, convert to list for handling.
    if not isinstance(files, list):
        files = [files]

    # Run through each file and read the data in
    for f in files:
        # Read in the first line of the file which has site, lat, lon, etc...
        df = str(pd.read_table(f, nrows=0).columns[0]).split(' ')
        ctr = 0
        for d in df:
            if len(d) > 0:
                if d == 'lat' or d == 'lon':
                    data[h1_names[ctr]]['data'].append(float(d) / 100.0)
                else:
                    data[h1_names[ctr]]['data'].append(d)
                ctr += 1

        # Set counts and errors
        error = False
        error_ct = 0
        ct = 0
        # Loop through while there's no errors i.e. eof
        while error is False:
            try:
                # Read in the initial headers to get information used to parse data
                if ct == 0:
                    df = str(pd.read_table(f, nrows=0, skiprows=[0]).columns[0]).split(' ')
                    ctr = 0
                    for d in df:
                        if len(d) > 0:
                            data[h2_names[ctr]]['data'].append(d)
                            ctr += 1
                    # Read in third row of header information
                    df = str(pd.read_table(f, nrows=0, skiprows=[0, 1]).columns[0]).split(' ')
                    ctr = 0
                    for d in df:
                        if len(d) > 0:
                            data[h3_names[ctr]]['data'].append(d)
                            ctr += 1
                    # Read in the data based on number of gates
                    df = pd.read_csv(
                        f,
                        skiprows=[0, 1, 2],
                        nrows=int(data['n_gates']['data'][-1]) - 1,
                        sep=r'\s+',
                        names=list(names.keys()),
                    )
                    index2 = 0
                else:
                    # Set indices for parsing data, reading 2 headers and then the columns of data
                    index1 = ct * int(data['n_gates']['data'][-1])
                    index2 = index1 + int(data['n_gates']['data'][-1]) + 2 * ct + 4
                    df = str(
                        pd.read_table(f, nrows=0, skiprows=list(range(index2 - 1))).columns[0]
                    ).split(' ')
                    ctr = 0
                    for d in df:
                        if len(d) > 0:
                            data[h2_names[ctr]]['data'].append(d)
                            ctr += 1
                    df = str(
                        pd.read_table(f, nrows=0, skiprows=list(range(index2))).columns[0]
                    ).split(' ')
                    ctr = 0
                    for d in df:
                        if len(d) > 0:
                            data[h3_names[ctr]]['data'].append(d)
                            ctr += 1
                    df = pd.read_csv(
                        f,
                        skiprows=list(range(index2 + 1)),
                        nrows=int(data['n_gates']['data'][-1]) - 1,
                        sep=r'\s+',
                        names=list(names.keys()),
                    )

                # Add data from the columns to the dictionary
                for n in names:
                    data[n]['data'].append(df[n].to_list())

                # Calculate the range based on number of gates, range to first gate and range between gates
                if len(data['range']['data']) == 0:
                    ranges = float(data['first_range_gate']['data'][-1]) + np.array(
                        range(int(data['n_gates']['data'][-1]) - 1)
                    ) * float(data['range_between_gates']['data'][-1])
                    data['range']['data'] = ranges

                # Calculate a time
                time = dt.datetime(
                    int('20' + data['year']['data'][-1]),
                    1,
                    1,
                    int(data['hour']['data'][-1]),
                    int(data['minute']['data'][-1]),
                    int(data['second']['data'][-1]),
                ) + dt.timedelta(days=int(data['day_of_year']['data'][-1]) - 1)
                data['time']['data'].append(time)

                # Range correct the snr which converts it essentially to an uncalibrated reflectivity
                snr_rc = data['snr']['data'][-1] - 20.0 * np.log10(1.0 / (ranges / 1000.0) ** 2)
                data['reflectivity_uncalibrated']['data'].append(snr_rc)
            except Exception as e:
                # Handle errors, if end of file then continue on, if something else
                # try the next block of data but if it errors another time in this file move on
                if isinstance(e, pd.errors.EmptyDataError) or error_ct > 1:
                    error = True
                else:
                    print(e)
                    pass
                error_ct += 1
            ct += 1

    # Convert dictionary to Dataset
    ds = xr.Dataset().from_dict(data)

    return ds



"""
Investigate Temperature Quantiles
---------------------------------

Investigate Temperature Quantiles
using DistributionDisplay Violin Plots

Written: Joe O'Brien

"""

import matplotlib.pyplot as plt
from arm_test_data import DATASETS

import act
from act.io.icartt import read_icartt

# Call the read_icartt function, which supports input
# for ICARTT (v2.0) formatted files.
# Example file is ARM Aerial Facility Navigation Data
filename_icartt = DATASETS.fetch('AAFNAV_COR_20181104_R0.ict')
ds = read_icartt(filename_icartt)

# Create a DistributionDisplay object to compare fields
display = act.plotting.DistributionDisplay(ds)

# Compare aircraft ground speed with ambient temperature
display.plot_violin(
    'ambient_temp',
    positions=[1.0],
)

display.plot_violin(
    'total_temp',
    positions=[2.0],
    set_title='Aircraft Temperatures 2018-11-04',
)

# Update the tick information
display.axes[0].set_xticks([0.5, 1, 2, 2.5])
ticks = ['', 'Ambient Air\nTemp', 'Total\nTemperature', '']
display.axes[0].set_xticklabels(ticks)

# Update the y-axis label
display.axes[0].set_ylabel('Temperature Observations [C]')
plt.show()



"""
Query and plot ASOS data
===========================================

This example shows how to plot timeseries of ASOS data from
Chicago O'Hare airport.

"""

from datetime import datetime

import matplotlib.pyplot as plt

import act

time_window = [datetime(2020, 2, 4, 2, 0), datetime(2020, 2, 10, 10, 0)]
my_asoses = act.discovery.get_asos_data(time_window, station='ORD', regions='IL')

display = act.plotting.TimeSeriesDisplay(my_asoses['ORD'], subplot_shape=(2,), figsize=(15, 10))
display.plot('temp', subplot_index=(0,))
display.plot_barbs_from_u_v(u_field='u', v_field='v', subplot_index=(1,))
display.axes[1].set_ylim([0, 2])
plt.show()



"""
Planetary Boundary Layer Height Gradient Method Retrievals
----------------------------------------------------------

This example shows how to estimate the planetary boundary layer
height via a gradient method retrieval

Author: Joe O'Brien
"""

from arm_test_data import DATASETS

import act

# Read Ceilometer data for an example
filename_ceil = DATASETS.fetch('sgpceilC1.b1.20190101.000000.nc')
ds = act.io.arm.read_arm_netcdf(filename_ceil)

# Apply corrections to the dataset
ds = act.corrections.correct_ceil(ds, var_name='backscatter')

# Estimate PBL Height via a gradient method
ds = act.retrievals.pbl_lidar.calculate_gradient_pbl(ds, parm="backscatter", smooth_dis=3)

# Estimate PBL Height via a modified gradient method
ds = act.retrievals.pbl_lidar.calculate_modified_gradient_pbl(
    ds, parm="backscatter", threshold=1e-4, smooth_dis=3
)

# Plot the pbl height estimates
display = act.plotting.TimeSeriesDisplay(ds, subplot_shape=(2,), figsize=(10, 8))

# plot the CL backscatter before overlaying the Gradient Method PBL Height
display.plot(
    'backscatter',
    subplot_index=(0,),
    cmap='ChaseSpectral',
    vmin=-6,
    vmax=6,
    set_title='SGP Ceilometer PBL Height Estimate via Gradient Method',
)

# overlay the PBL Height estimate, compute ~10min temporal averages
display.axes[0].plot(
    ds['time'].values,
    ds['pbl_gradient'].rolling(time=38, min_periods=3, center=True).mean().values,
    color='k',
)
# shorten the range
display.set_yrng([0, 2000], subplot_index=(0,))

# plot the CL backscatter before overlaying the Modified Gradient PBL Height
display.plot(
    'backscatter',
    subplot_index=(1,),
    cmap='ChaseSpectral',
    vmin=-6,
    vmax=6,
    set_title='SGP Ceilometer PBL Height Estimate via Modified Gradient Method',
)

# overlay the PBL Height estimate, compute ~10min temporal averages
display.axes[1].plot(
    ds['time'].values,
    ds['pbl_mod_gradient'].rolling(time=38, min_periods=3, center=True).mean().values,
    color='k',
)
# shorten the range
display.set_yrng([0, 2000], subplot_index=(1,))



"""
Script for downloading data from Ameriflux's Data Webservice

"""

import os
import warnings

import requests

warnings.simplefilter('always')


def download_ameriflux_data(
    user_id,
    user_email,
    site_ids,
    data_product="FLUXNET",
    data_policy=None,
    data_variant="FULLSET",
    agree_policy=True,
    intended_use=None,
    description=None,
    out_dir=None,
    **kwargs,
):
    """
    This tool will allows users to download Ameriflux data. This code is based on the
    original R code found here: https://github.com/chuhousen/amerifluxr

    Parameters
    ----------
    user_id : str
        The user's Ameriflux user ID.
    user_email : str
        The user's email address.
    site_ids : list
        A list of valid site_ids to download data from. List of available sites can be found here:
        https://ameriflux.lbl.gov/sites/site-search/
    data_product : str
        Data product type. Options are BASE-BADM, BIF, or FLUXNET.
        Default is FLUXNET. For more on data products:
        https://ameriflux.lbl.gov/data/flux-data-products/
        https://ameriflux.lbl.gov/data/badm/
    data_policy : str
        Data policy under which data has been licensed by the PI. Options are CCBY4.0 or LEGACY.
    data_variant : str or None
       Variant used for FLUXNET data product. Options are SUBSET, FULLSET, or None.
       Default is FUllSET
    agree_policy : bool
        Acknowledge you read and agree to the AmeriFlux Data use policy. Data policy can be found here:
        https://ameriflux.lbl.gov/data/data-policy/
    intended_use : str
         Planned use for data downloaded. Select best match. Shorter code options are available'
         and when provided will be used to provide full intended use, examples of user codes and what they
         correlate to:
         "synthesis": "Research - Multi-site synthesis",
         "remote_sensing": "Research - Remote sensing",
         "model": "Research - Land model/Earth system model",
         "other_research": "Research - Other",
         "education": "Education (Teacher or Student)",
         "other": "Other"
    description : str
        Brief description of intended use. This will be recorded in the data download log and emailed to site’s PI.
    out_dir : str
        The output directory for the data. Set to None to make a folder in the
        current working directory with the same name as *datastream* to place
        the files in.

    Notes
    -----
    This programmatic interface allows users to query and automate
    machine-to-machine downloads of Ameriflux data. This tool uses a REST URL and
    specific parameters mentioned above and data files matching
    the criteria will be returned to the user and downloaded.

    To login/register for an Ameriflux account:
    https://amfcdn.lbl.gov/create

    Examples
    --------
    This code will download a zip file for BASE-BADM data product at site
    US-A37. See the Notes for information on how to obtain a username and token.

    .. code-block:: python

        act.discovery.download_ameriflux_data(
            user_id, user_email, data_product="BASE-BADM", data_policy="CCBY4.0", data_variant="FULLSET",
            site_ids=["US-A37"], agree_policy=True, intended_use="synthesis",
            description="I intend to use this data for research", out_dir="/home/user/ameriflux_data/",
        )

    Returns
    -------
    files : list
        Returns list of files retrieved

    """
    # Check all inputs are valid
    if not isinstance(user_id, str):
        raise ValueError("user_id should be a string...")

    if not isinstance(user_email, str) or "@" not in user_email:
        raise ValueError("user_email not a valid email...")

    # Check if site_id are valid site IDs
    check_id = _check_site_id(site_ids)
    if isinstance(site_ids, list):
        if any(not valid for valid in check_id):
            warnings.warn(
                f"{', '.join([site_ids[i] for i, valid in enumerate(check_id) if not valid])} not valid AmeriFlux Site ID",
                UserWarning,
            )
            site_ids = [site_ids[i] for i, valid in enumerate(check_id) if valid]
    elif isinstance(site_ids, str):
        if not (check_id or site_ids in ["AA-Flx", "AA-Net"]):
            site_ids = None

    if not site_ids:
        raise ValueError("No valid Site ID in site_ids...")

    # Check if intended use is valid
    intended_use_list = [
        "synthesis",
        "remote_sensing",
        "model",
        "other_research",
        "education",
        "other",
    ]

    if intended_use not in intended_use_list:
        raise ValueError("Invalid intended_use input...")

    # Check if out_dir is reachable
    if out_dir is None:
        os.makedirs(os.getcwd() + '/data/')
        out_dir = os.getcwd() + '/data/'
    else:
        if not os.path.isdir(out_dir):
            os.makedirs(out_dir)

    # Prompt for data policy agreement
    if data_policy == "CCBY4.0":
        warnings.warn(
            "\n"
            "Data use guidelines for AmeriFlux CC-BY-4.0 Data Policy:\n"
            "(1) Data user is free to Share (copy and redistribute the material in any medium or format) and/or Adapt (remix, transform, and build upon the material) for any purpose.\n"
            "(2) Provide a citation to each site data product that includes the data-product DOI and/or recommended publication.\n"
            "(3) Acknowledge funding for supporting AmeriFlux data portal: U.S. Department of Energy Office of Science.\n"
            "\n",
            PolicyWarning,
        )
    elif data_policy == "LEGACY":
        warnings.warn(
            "\n"
            "Data use guidelines for AmeriFlux LEGACY License:\n"
            "(1) When you start in-depth analysis that may result in a publication, contact the data contributors directly, so that they have the opportunity to contribute substantively and become a co-author.\n"
            "(2) Provide a citation to each site data product that includes the data-product DOI.\n"
            "(3) Acknowledge funding for site support if it was provided in the data download information.\n"
            "(4) Acknowledge funding for supporting AmeriFlux data portal: U.S. Department of Energy Office of Science.\n"
            "\n",
            PolicyWarning,
        )
    else:
        raise ValueError("Specify a valid data policy before proceeding...")

    if not agree_policy:
        raise ValueError("Acknowledge data policy before proceeding...")

    if "is_test" in kwargs:
        if kwargs['is_test'] is True:
            test_key = "true"
        else:
            test_key = ""
    else:
        test_key = ""

    # Payload for download web service
    params = {
        "user_id": user_id,
        "user_email": user_email,
        "data_policy": data_policy,
        "data_product": data_product,
        "data_variant": data_variant,
        "site_ids": site_ids,
        "intended_use": intended_use,
        "description": f"{description} [Atmospheric data Community Toolkit download]",
        "is_test": test_key,
    }

    result = requests.post(
        _ameriflux_endpoints("data_download"),
        json=params,
        headers={
            "content-type": "application/json",
            "accept": "applicaton/json",
        },
    )
    # Check if FTP returns correctly
    if result.status_code == 200:
        link = result.json()
        ftplink = [data_url['url'] for data_url in link.get('data_urls', [])]

        # Check if any site_id has no data
        if not ftplink:
            raise ValueError(f"Cannot find data from {site_ids}")

        # Avoid downloading fluxnet_bif for now
        if (
            isinstance(site_ids, str)
            and site_ids == "AA-Flx"
            and data_policy == "CCBY4.0"
            and len(ftplink) > 1
        ):
            ftplink = [url for url in ftplink if "FLUXNET-BIF" not in url]

        # Get zip file names
        outfname = [os.path.basename(url).split("?")[0] for url in ftplink]

        # Check if any site_ids has no data
        if len(outfname) < len(site_ids):
            miss_site_id = [
                sid for sid in site_ids if sid not in [fname[4:10] for fname in outfname]
            ]
            warnings.warn(f"Cannot find data from {miss_site_id}")

        # Download sequentially
        output_zip_file = [os.path.join(out_dir, fname) for fname in outfname]
        for i, url in enumerate(ftplink):
            response = requests.get(url, stream=True)
            with open(output_zip_file[i], 'wb') as f:
                for chunk in response.iter_content(chunk_size=8192):
                    f.write(chunk)

        # Check if downloaded files exist
        miss_download = [i for i, file in enumerate(output_zip_file) if not os.path.exists(file)]
        if miss_download:
            warnings.warn(
                f"Cannot download {[output_zip_file[i] for i in miss_download]} from {[ftplink[i] for i in miss_download]}"
            )

    else:
        raise ValueError("Data download fails, timeout or server error...")

    return output_zip_file


# Return AmeriFlux server endpoints
def _ameriflux_endpoints(endpoint="sitemap"):
    """Retrieves urls for different ameriflux server endpoints. Options include"
    sitemap, site_ccby4, data_year, data_download, and variables"""
    # base urls
    base_url = "https://amfcdn.lbl.gov/api/v2/"

    # what to return
    url = {
        "sitemap": base_url + "site_info_display/AmeriFlux",
        "site_ccby4": base_url + "site_availability/AmeriFlux/BIF/CCBY4.0",
        "data_download": base_url + "data_download",
    }.get(endpoint)
    return url


def _check_site_id(x):
    """Checks if user provided site_ids are valid."""
    headers = {
        'Content-Type': 'application/json',
        'user-agent': 'Mozilla/5.0 (X11; Linux x86_64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/67.0.3396.99 Safari/537.36',
    }
    response = requests.get(_ameriflux_endpoints("sitemap"), headers=headers)
    response_json = response.json()
    site_ids = [i['site_id'] for i in response_json['values']]
    chk_id = [site_id in site_ids for site_id in x]
    return chk_id


class PolicyWarning(UserWarning):
    pass



"""
Configuration file for the Python Atmospheric data Community Toolkit (ACT)
The values for a number of ACT parameters and the default metadata created
when reading files, correcting fields, etc. is controlled by this single
Python configuration file.

Examples:
---------
from act.config import DEFAULT_DATASTREAM_NAME

"""

DEFAULT_DATASTREAM_NAME = 'act_datastream'



"""
This module contains I/O operations for the U.S. Department of Energy
AmeriFlux program (https://ameriflux.lbl.gov/).
"""

import re
import warnings
from collections import defaultdict

import numpy as np
import pandas as pd
import xarray as xr


def read_ameriflux(
    filename,
    metadata_filename=None,
    data_type=None,
    timestep=None,
    rename_vars_dict=None,
    variable_units_dict=None,
):
    """
    Returns `xarray.Dataset` with stored data from FLUXNET or BASE-BADM and
    possible metadata from a user-defined metadata excel files for a single datastream.

    Parameters
    ----------
    filename : str
        Filename of Ameriflux dataset
    metadata_filename : str
        Excel file usually provided with Ameriflux dataset that contains the data's metadata.
        Default is None.
    data_type : str
        Type of data file to be read. Valid options are 'fluxnet' and 'base'.
        Default is None and will try to determine data type from filename.
    timestep : str
        Timestep of data, this parameter is only used for 'fluxnet' data types and if
        the time format can't be determined by the filename.
        Default is None, options are 'year', 'month', 'week', 'day', and 'hour'.
    rename_vars_dict : dict
        A dictionary containing current variable names and new variable names to replace
        the current variable names.
        Default is None and variables are not renamed.
    variable_units_dict : dict
        A dictionary containing current variable names and units to be added to that variable.
        Default is None and uses current units from Ameriflux's unit database.

    Returns
    -------
    ds : xarray.Dataset
        ACT Xarray dataset.

    """
    default_units_dict = {
        'COND_WATER': 'S cm-1',
        'DO': 'mol L-1',
        'PCH4': 'nmolCH4 mol-1',
        'PCO2': 'molCO2 mol-1',
        'PN2O': 'nmolN2O mol-1',
        'PPFD_UW_IN': 'molPhotons m-2 s-1',
        'TW': 'deg C',
        'DBH': 'cm',
        'LEAF_WET': '%',
        'SAP_DT': 'deg C',
        'SAP_FLOW': 'mmolH2O m-2 s-1',
        'T_BOLE': 'deg C',
        'T_CANOPY': 'deg C',
        'FETCH_70': 'm',
        'FETCH_80': 'm',
        'FETCH_90': 'm',
        'FETCH_FILTER': 'nondimensional',
        'FETCH_MAX': 'm',
        'CH4': 'nmolCH4 mol-1',
        'CH4_MIXING_RATIO': 'nmolCH4 mol-1',
        'CO': 'nmolCO mol-1',
        'CO2': 'molCO2 mol-1',
        'CO2_MIXING_RATIO': 'molCO2 mol-1',
        'CO2_SIGMA': 'molCO2 mol-1',
        'CO2C13': '(permil)',
        'FC': 'molCO2 m-2 s-1',
        'FCH4': 'nmolCH4 m-2 s-1',
        'FN2O': 'nmolN2O m-2 s-1',
        'FNO': 'nmolNO m-2 s-1',
        'FNO2': 'nmolNO2 m-2 s-1',
        'FO3': 'nmolO3 m-2 s-1',
        'H2O': 'mmolH2O mol-1',
        'H2O_MIXING_RATIO': 'mmolH2O mol-1',
        'H2O_SIGMA': 'mmolH2O mol-1',
        'N2O': 'nmolN2O mol-1',
        'N2O_MIXING_RATIO': 'nmolN2O mol-1',
        'NO': 'nmolNO mol-1',
        'NO2': 'nmolNO2 mol-1',
        'O3': 'nmolO3 mol-1',
        'SC': 'molCO2 m-2 s-1',
        'SCH4': 'nmolCH4 m-2 s-1',
        'SN2O': 'nmolN2O m-2 s-1',
        'SNO': 'nmolNO m-2 s-1',
        'SNO2': 'nmolNO2 m-2 s-1',
        'SO2': 'nmolSO2 mol-1',
        'SO3': 'nmolO3 m-2 s-1',
        'FH2O': 'mmolH2O m-2 s-1',
        'G': 'W m-2',
        'H': 'W m-2',
        'LE': 'W m-2',
        'SB': 'W m-2',
        'SG': 'W m-2',
        'SH': 'W m-2',
        'SLE': 'W m-2',
        'PA': 'kPa',
        'PBLH': 'm',
        'RH': '%',
        'T_SONIC': 'deg C',
        'T_SONIC_SIGMA': 'deg C',
        'TA': 'deg C',
        'VPD': 'hPa',
        'D_SNOW': 'cm',
        'P': 'mm',
        'P_RAIN': 'mm',
        'P_SNOW': 'mm',
        'RUNOFF': 'mm',
        'STEMFLOW': 'mm',
        'THROUGHFALL': 'mm',
        'ALB': '%',
        'APAR': 'molPhoton m-2 s-1',
        'EVI': 'nondimensional',
        'FAPAR': '%',
        'FIPAR': '%',
        'LW_BC_IN': 'W m-2',
        'LW_BC_OUT': 'W m-2',
        'LW_IN': 'W m-2',
        'LW_OUT': 'W m-2',
        'MCRI': 'nondimensional',
        'MTCI': 'nondimensional',
        'NDVI': 'nondimensional',
        'NETRAD': 'W m-2',
        'NIRV': 'W m-2 sr-1 nm-1',
        'PPFD_BC_IN': 'molPhoton m-2 s-1',
        'PPFD_BC_OUT': 'molPhoton m-2 s-1',
        'PPFD_DIF': 'molPhoton m-2 s-1',
        'PPFD_DIR': 'molPhoton m-2 s-1',
        'PPFD_IN': 'molPhoton m-2 s-1',
        'PPFD_OUT': 'molPhoton m-2 s-1',
        'PRI': 'nondimensional',
        'R_UVA': 'W m-2',
        'R_UVB': 'W m-2',
        'REDCI': 'nondimensional',
        'REP': 'nm',
        'SPEC_NIR_IN': 'W m-2 nm-1',
        'SPEC_NIR_OUT': 'W m-2 sr-1 nm-1',
        'SPEC_NIR_REFL': 'nondimensional',
        'SPEC_PRI_REF_IN': 'W m-2 nm-1',
        'SPEC_PRI_REF_OUT': 'W m-2 sr-1 nm-1',
        'SPEC_PRI_REF_REFL': 'nondimensional',
        'SPEC_PRI_TGT_IN': 'W m-2 nm-1',
        'SPEC_PRI_TGT_OUT': 'W m-2 sr-1 nm-1',
        'SPEC_PRI_TGT_REFL': 'nondimensional',
        'SPEC_RED_IN': 'W m-2 nm-1',
        'SPEC_RED_OUT': 'W m-2 sr-1 nm-1',
        'SPEC_RED_REFL': 'nondimensional',
        'SR': 'nondimensional',
        'SW_BC_IN': 'W m-2',
        'SW_BC_OUT': 'W m-2',
        'SW_DIF': 'W m-2',
        'SW_DIR': 'W m-2',
        'SW_IN': 'W m-2',
        'SW_OUT': 'W m-2',
        'TCARI': 'nondimensional',
        'SWC': '%',
        'SWP': 'kPa',
        'TS': 'deg C',
        'TSN': 'deg C',
        'WTD': 'm',
        'MO_LENGTH': 'm',
        'TAU': 'kg m-1 s-2',
        'U_SIGMA': 'm s-1',
        'USTAR': 'm s-1',
        'V_SIGMA': 'm s-1',
        'W_SIGMA': 'm s-1',
        'WD': 'Decimal degrees',
        'WD_SIGMA': 'decimal degree',
        'WS': 'm s-1',
        'WS_MAX': 'm s-1',
        'ZL': 'nondimensional',
        'GPP': 'molCO2 m-2 s-1',
        'NEE': 'molCO2 m-2 s-1',
        'RECO': 'molCO2 m-2 s-1',
        'FC_SSITC_TEST': 'nondimensional',
        'FCH4_SSITC_TEST': 'nondimensional',
        'FN2O_SSITC_TEST': 'nondimensional',
        'FNO_SSITC_TEST': 'nondimensional',
        'FNO2_SSITC_TEST': 'nondimensional',
        'FO3_SSITC_TEST': 'nondimensional',
        'H_SSITC_TEST': 'nondimensional',
        'LE_SSITC_TEST': 'nondimensional',
        'TAU_SSITC_TEST': 'nondimensional',
    }
    # Reader section for BASE BADM datasets
    # Differs from fluxnet as there is metadata in the first few lines
    # of the csv file.
    if data_type is not None:
        if data_type.lower() == 'base':
            file_data_type = 'base'
        elif data_type.lower() == 'fluxnet':
            file_data_type = 'fluxnet'
        else:
            raise ValueError("Must choose a valid data_type, either 'base' or 'fluxnet'")
    # Try to automatically get data type if data type isn't provided
    elif data_type is None and 'FLUXNET' in filename:
        file_data_type = 'fluxnet'
    elif data_type is None and 'BASE-BADM' in filename:
        file_data_type = 'base'
    else:
        raise ValueError(
            'Could not determine the data type from the filename '
            ' Please provide a valid data type to the data_type parameter!'
        )

    if file_data_type == 'base':
        # Grab site and version metadata
        metadata = pd.read_csv(filename, header=None, nrows=2, sep=':', index_col=0)
        site = metadata.loc['# Site']
        version = metadata.loc['# Version']

        # Files are hourly, set the time format
        _format = "%Y%m%d%H%M"

        # Read in actual data
        df = pd.read_csv(filename, skiprows=2)

        # Set time and format time to datetime64
        key = 'TIMESTAMP_START'
        df['time'] = pd.to_datetime(df[key], format=_format)
        df = df.set_index('time')
        df = df.drop(columns=['TIMESTAMP_START', 'TIMESTAMP_END'])

        # Convert dataframe to xarray dataset
        ds = xr.Dataset.from_dataframe(df)

        # Add site and version to attributes
        ds.attrs['site'] = site.values[0].strip()
        ds.attrs['version'] = version.values[0].strip()

    # Reader for fluxnet files
    # Fluxnet files can be formatted in different time samplings
    elif file_data_type == 'fluxnet':
        # Checks timestep in filename, if not, will check timestep parameter
        if 'YY' in filename or timestep == 'year':
            _format = "%Y"
        elif 'MM' in filename or timestep == 'month':
            _format = "%Y%m"
        elif 'DD' in filename or 'WW' in filename or timestep == 'week' or timestep == 'day':
            _format = "%Y%m%d"
        elif 'HH' in filename or timestep == 'day':
            _format = "%Y%m%d%H%M"
        else:
            raise ValueError(
                "Incorrect timestep provided or no timestep determined from filename, "
                "please provide either year, month, week, day or hour for the timestep parameter."
            )
        # Read data into a pandas dataframe
        df = pd.read_csv(filename)

        # Set time and convert to datetime64
        if 'TIMESTAMP_START' in df:
            key = 'TIMESTAMP_START'
            df['time'] = pd.to_datetime(df[key], format=_format)
            df = df.set_index('time')
            df = df.drop(columns=['TIMESTAMP_START', 'TIMESTAMP_END'])
        else:
            key = 'TIMESTAMP'
            df['time'] = pd.to_datetime(df[key], format=_format)
            df = df.set_index('time')
            df = df.drop(columns=[key])

        # Convert to xarray dataset
        ds = xr.Dataset.from_dataframe(df)

    # Renames variables if user provides new names
    if rename_vars_dict is not None:
        ds = ds.rename_vars(rename_vars_dict)

    # Add _FILL_VALUE attribute
    for key in ds.variables.keys():
        if -9999.0 in ds.variables[key].values or -9999 in ds.variables[key].values:
            ds.variables[key].attrs['_FILL_VALUE'] = -9999

    # Add nans where fill value is present
    ds = ds.where(ds != -9999.0)

    # Add units from the unit dictionary
    # Matches keys that have different levels as well such as SWC_1_1_1
    # Only works for base dataset
    if variable_units_dict is None and file_data_type == 'base':
        for key in ds.variables.keys():
            try:
                if re.match(r'TS_[\d]', key):
                    unit_key = 'TS'
                    ds.variables[key].attrs['units'] = default_units_dict[unit_key]
                elif re.match(r'SWC_[\d]', key):
                    unit_key = 'SWC'
                    ds.variables[key].attrs['units'] = default_units_dict[unit_key]
                elif re.match(r'G_[\d]', key):
                    unit_key = 'G'
                    ds.variables[key].attrs['units'] = default_units_dict[unit_key]
                elif re.match(r'VPD_', key):
                    unit_key = 'VPD'
                    ds.variables[key].attrs['units'] = default_units_dict[unit_key]
                else:
                    ds.variables[key].attrs['units'] = default_units_dict[key]
            except KeyError:
                continue

    if variable_units_dict is not None:
        for key in ds.variables.keys():
            ds.variables[key].attrs['units'] = variable_units_dict[key]

    # Adds metadata to the dataset if a metadata file is provided
    if metadata_filename is not None:
        ds = _ameriflux_metadata_processing(ds, metadata_filename)

    return ds


def _ameriflux_metadata_processing(ds, metadata_filename):
    """Adds metadata to an ameriflux dataset if a metadata file is provided."""
    # Read in metadata excel file if provided.
    meta_df = pd.read_excel(metadata_filename)

    # Create a list for attributes and their respective values
    meta_key_list = meta_df['VARIABLE'].to_list()
    meta_value_list = meta_df['DATAVALUE'].to_list()

    # Add attrs that are non duplicates as duplicates require more processing below
    non_duplicates = [item for item in meta_key_list if meta_key_list.count(item) == 1]
    data_dict = defaultdict(list)
    for i, j in zip(meta_key_list, meta_value_list):
        data_dict[i].append(j)
    for key in non_duplicates:
        ds.attrs[key] = data_dict[key][0]

    # The code below to retrieve group metadata was done so that
    # If the order changes of these attributes, which it can, it will find them regardless
    # Also checks if a specific attribute is missing for a team member but not others
    meta_arr = np.array(meta_key_list)
    # Retrieve team_member metadata
    team_indices = np.where(meta_arr == 'TEAM_MEMBER_NAME')[0]
    team_members = []
    valid_names = [
        'TEAM_MEMBER_NAME',
        'TEAM_MEMBER_EMAIL',
        'TEAM_MEMBER_INSTITUTION',
        'TEAM_MEMBER_ROLE',
        'TEAM_MEMBER_ORCID',
        'TEAM_MEMBER_ADDRESS',
    ]
    for i, j in enumerate(team_indices):
        if j == team_indices[-1]:
            team_info_range = np.arange(team_indices[i], team_indices[i] + 5)
            team_members_str = []
            [
                team_members_str.append(str(meta_key_list[k]) + ':' + str(meta_value_list[k]))
                for k in team_info_range
                if meta_key_list[k] in valid_names
            ]
            result_str = ", ".join(team_members_str)
            team_members.append(result_str)
            break
        else:
            team_info_range = np.arange(team_indices[i], team_indices[i + 1])
            team_members_str = []
            [
                team_members_str.append(str(meta_key_list[k]) + ':' + str(meta_value_list[k]))
                for k in team_info_range
                if meta_key_list[k] in valid_names
            ]
            result_str = ", ".join(team_members_str)
            team_members.append(result_str)
    ds.attrs['TEAM_MEMBERS'] = team_members

    # Retrieve doi contributor metadata
    doi_indices = np.where(meta_arr == 'DOI_CONTRIBUTOR_DATAPRODUCT')[0]
    doi_members = []
    valid_doi_names = [
        'DOI_CONTRIBUTOR_DATAPRODUCT',
        'DOI_CONTRIBUTOR_NAME',
        'DOI_CONTRIBUTOR_ROLE',
        'DOI_CONTRIBUTOR_INSTITUTION',
        'DOI_CONTRIBUTOR_EMAIL',
        'DOI_CONTRIBUTOR_ORCID',
        'DOI_CONTRIBUTOR_DATE_START',
        'DOI_CONTRIBUTOR_DATE_END',
        'DOI CONTRIBUTOR_ADDRESS',
    ]
    for i, j in enumerate(doi_indices):
        if j == doi_indices[-1]:
            doi_info_range = np.arange(doi_indices[i], doi_indices[i] + 5)
            doi_contrib_str = []
            [
                doi_contrib_str.append(str(meta_key_list[k]) + ':' + str(meta_value_list[k]))
                for k in doi_info_range
                if meta_key_list[k] in valid_doi_names
            ]
            result_doi_str = ", ".join(doi_contrib_str)
            doi_members.append(result_doi_str)
            break
        else:
            doi_info_range = np.arange(doi_indices[i], doi_indices[i + 1])
            doi_contrib_str = []
            [
                doi_contrib_str.append(str(meta_key_list[k]) + ':' + str(meta_value_list[k]))
                for k in doi_info_range
                if meta_key_list[k] in valid_doi_names
            ]
            result_doi_str = ", ".join(doi_contrib_str)
            doi_members.append(result_doi_str)
    ds.attrs['DOI_CONTRIBUTORS'] = doi_members

    # Retrieve flux method metadata
    flux_indices = np.where(meta_arr == 'FLUX_MEASUREMENTS_METHOD')[0]
    flux_members = []
    valid_flux_names = [
        'FLUX_MEASUREMENTS_METHOD',
        'FLUX_MEASUREMENTS_VARIABLE',
        'FLUX_MEASUREMENTS_DATE_START',
        'FLUX_MEASUREMENTS_DATE_END',
        'FLUX_MEASUREMENTS_OPERATIONS',
    ]
    for i, j in enumerate(flux_indices):
        if j == flux_indices[-1]:
            flux_info_range = np.arange(flux_indices[i], flux_indices[i] + 5)
            flux_str = []
            [
                flux_str.append(str(meta_key_list[k]) + ':' + str(meta_value_list[k]))
                for k in flux_info_range
                if meta_key_list[k] in valid_flux_names
            ]
            result_flux_str = ", ".join(flux_str)
            flux_members.append(result_flux_str)
            break
        else:
            flux_info_range = np.arange(flux_indices[i], flux_indices[i + 1])
            flux_str = []
            [
                flux_str.append(str(meta_key_list[k]) + ':' + str(meta_value_list[k]))
                for k in flux_info_range
                if meta_key_list[k] in valid_flux_names
            ]
            result_flux_str = ", ".join(flux_str)
            flux_members.append(result_flux_str)
    ds.attrs['FLUX_MEASUREMENTS_METHODS'] = flux_members
    return ds


def convert_to_ameriflux(
    ds,
    variable_mapping=None,
    soil_mapping=None,
    depth_profile=[2.5, 5, 10, 15, 20, 30, 35, 50, 75, 100],
    include_missing_variables=False,
    **kwargs,
):
    """
    Returns `xarray.Dataset` with stored data and metadata from a user-defined
    query of ARM-standard netCDF files from a single datastream. Has some procedures
    to ensure time is correctly formatted in returned Dataset.

    Parameters
    ----------
    ds : xarray.Dataset
        Dataset of data to convert to AmeriFlux format
    variable_mapping : dict
        Dictionary of variables mappings. The key should be the name of the variable
        in the Dataset with the values being dictionaries of the AmeriFlux name and units.
        For example::

            var_mapping = {
                'co2_flux': {'name': 'FC', 'units': 'umol/(m^2 s)'},
            }

    soil_mapping : dict
        Dictionary of soil variables mappings following the same formatting as variable_mapping.
        It is understood that the AmeriFlux name may be the same for some variables. This
        script attempts to automatically name these measurements. If a variable is not dimensioned
        by a depth nor has a sensor_height attribute, it will automatically assume that it's
        at the first depth in the depth_profile variable.
    depth_profile : list
        List of depths that the variables will be mapped to. If a depth is not in this list,
        the index chosen will be the one closest to the depth value.
    include_missing_variables : bool
        If there are variables that are completely missing (-9999) choose whether or not to
        include them in the DataFrame.

    Returns
    -------
    df : pandas.DataFrame or None
        Returns a pandas dataframe for easy writing to csv.
    """
    # Use ARM variable mappings if none provided
    if variable_mapping is None:
        warnings.warn('Variable mapping was not provided, using default ARM mapping')
        # Define variable mapping and units
        # The key is the variable name in the data and the name in the dictionary
        # is the AmeriFlux Name
        var_mapping = {
            'co2_flux': {'name': 'FC', 'units': 'umol/(m^2 s)'},
            'co2_molar_fraction': {'name': 'CO2', 'units': 'nmol/mol'},
            'co2_mixing_ratio': {'name': 'CO2_MIXING_RATIO', 'units': 'umol/mol'},
            'h2o_mole_fraction': {'name': 'H2O', 'units': 'mmol/mol'},
            'h2o_mixing_ratio': {'name': 'H2O_MIXING_RATIO', 'units': 'mmol/mol'},
            'ch4_mole_fraction': {'name': 'CH4', 'units': 'nmol/mol'},
            'ch4_mixing_ratio': {'name': 'CH4_MIXING_RATIO', 'units': 'nmol/mol'},
            'momentum_flux': {'name': 'TAU', 'units': 'kg/(m s^2)'},
            'sensible_heat_flux': {'name': 'H', 'units': 'W/m^2'},
            'latent_flux': {'name': 'LE', 'units': 'W/m^2'},
            'air_temperature': {'name': 'TA', 'units': 'deg C'},
            'air_pressure': {'name': 'PA', 'units': 'kPa'},
            'relative_humidity': {'name': 'RH', 'units': '%'},
            'sonic_temperature': {'name': 'T_SONIC', 'units': 'deg C'},
            'water_vapor_pressure_defecit': {'name': 'VPD', 'units': 'hPa'},
            'Monin_Obukhov_length': {'name': 'MO_LENGTH', 'units': 'm'},
            'Monin_Obukhov_stability_parameter': {'name': 'ZL', 'units': ''},
            'mean_wind': {'name': 'WS', 'units': 'm/s'},
            'wind_direction_from_north': {'name': 'WD', 'units': 'deg'},
            'friction_velocity': {'name': 'USTAR', 'units': 'm/s'},
            'maximum_instantaneous_wind_speed': {'name': 'WS_MAX', 'units': 'm/s'},
            'down_short_hemisp': {'name': 'SW_IN', 'units': 'W/m^2'},
            'up_short_hemisp': {'name': 'SW_OUT', 'units': 'W/m^2'},
            'down_long': {'name': 'LW_IN', 'units': 'W/m^2'},
            'up_long': {'name': 'LW_OUT', 'units': 'W/m^2'},
            'albedo': {'name': 'ALB', 'units': '%'},
            'net_radiation': {'name': 'NETRAD', 'units': 'W/m^2'},
            'par_inc': {'name': 'PPFD_IN', 'units': 'umol/(m^2 s)'},
            'par_ref': {'name': 'PPFD_OUT', 'units': 'umol/(m^2 s)'},
            'precip': {'name': 'P', 'units': 'mm'},
        }

    # Use ARM variable mappings if none provided
    # Similar to the above.  This has only been tested on the ARM
    # ECOR, SEBS, STAMP, and AMC combined.  The automated naming may
    # not work for all cases
    if soil_mapping is None:
        warnings.warn('Soil variable mapping was not provided, using default ARM mapping')
        soil_mapping = {
            'surface_soil_heat_flux': {'name': 'G', 'units': 'W/m^2'},
            'soil_temp': {'name': 'TS', 'units': 'deg C'},
            'temp': {'name': 'TS', 'units': 'deg C'},
            'soil_moisture': {'name': 'SWC', 'units': '%'},
            'soil_specific_water_content': {'name': 'SWC', 'units': '%'},
            'vwc': {'name': 'SWC', 'units': '%'},
        }

    # Loop through variables and update units to the AmeriFlux standard
    for v in ds:
        if v in var_mapping:
            ds = ds.utils.change_units(variables=v, desired_unit=var_mapping[v]['units'])

    # Get start/end time stamps
    ts_start = ds['time'].dt.strftime('%Y%m%d%H%M').values
    ts_end = [
        pd.to_datetime(t + np.timedelta64(30, 'm')).strftime('%Y%m%d%H%M')
        for t in ds['time'].values
    ]
    data = {}
    data['TIMESTAMP_START'] = ts_start
    data['TIMESTAMP_END'] = ts_end

    # Loop through the variables in the var mapping dictionary and add data to dictionary
    for v in var_mapping:
        if v in ds:
            if 'missing_value' not in ds[v].attrs:
                ds[v].attrs['missing_value'] = -9999
            if np.all(ds[v].isnull()):
                if include_missing_variables:
                    data[var_mapping[v]['name']] = ds[v].values
            else:
                data[var_mapping[v]['name']] = ds[v].values
        else:
            if include_missing_variables:
                data[var_mapping[v]['name']] = np.full(ds['time'].shape, -9999)

    # Automated naming for the soil variables
    # Again, this may not work for other cases.  Careful review is needed.
    prev_var = ''
    for var in soil_mapping:
        if soil_mapping[var]['name'] != prev_var:
            h = 1
            r = 1
            prev_var = soil_mapping[var]['name']
        soil_vars = [
            v2
            for v2 in list(ds)
            if (v2.startswith(var)) & ('std' not in v2) & ('qc' not in v2) & ('net' not in v2)
        ]
        for i, svar in enumerate(soil_vars):
            vert = 1
            if ('avg' in svar) | ('average' in svar):
                continue
            soil_data = ds[svar].values
            data_shape = soil_data.shape
            if len(data_shape) > 1:
                coords = ds[svar].coords
                depth_name = list(coords)[-1]
                depth_values = ds[depth_name].values
                for depth_ind in range(len(depth_values)):
                    soil_data_depth = soil_data[:, depth_ind]
                    vert = np.where(depth_profile == depth_values[depth_ind])[0][0] + 1
                    new_name = '_'.join([soil_mapping[var]['name'], str(h), str(vert), str(r)])
                    data[new_name] = soil_data_depth
            else:
                if 'sensor_height' in ds[svar].attrs:
                    sensor_ht = ds[svar].attrs['sensor_height'].split(' ')
                    depth = abs(float(sensor_ht[0]))
                    units = sensor_ht[1]
                    if units == 'cm':
                        vert = np.argmin(np.abs(np.array(depth_profile) - depth)) + 1
                new_name = '_'.join([soil_mapping[var]['name'], str(h), str(vert), str(r)])
                data[new_name] = soil_data
            h += 1

    # Convert dictionary to dataframe and return
    df = pd.DataFrame(data)
    df = df.fillna(-9999.0)

    return df



"""
Functions and methods for performing solar radiation tests taken from
the BSRN Global Network recommended QC tests, V2.0

https://bsrn.awi.de

"""

import warnings

import dask.array as da
import numpy as np
from scipy.constants import Stefan_Boltzmann

from act.utils.data_utils import convert_units
from act.utils.geo_utils import get_solar_azimuth_elevation


def _calculate_solar_parameters(ds, lat_name, lon_name, solar_constant=1360.8):
    """
    Function to calculate solar zenith angles and solar constant adjusted
    to Earth Sun distance

    Parameters
    ----------
    ds : xarray.Dataset
        Dataset containing location variables
    lat_name : str
        Variable name for latitude
    lon_name : str
        Variable name for longitude
    solar_constant : float
        Solar constant in W/m^2

    Returns
    -------
    Tuple containing (solar zenith angle array, solar constant scalar)

    """
    latitude = ds[lat_name].values
    if latitude.size > 1:
        latitude = latitude[0]
    longitude = ds[lon_name].values
    if longitude.size > 1:
        longitude = longitude[0]

    # Calculate solar parameters
    elevation, _, solar_distance = get_solar_azimuth_elevation(
        latitude=latitude, longitude=longitude, time=ds['time'].values
    )
    solar_distance = np.nanmean(solar_distance)
    Sa = solar_constant / solar_distance**2

    sza = 90.0 - elevation

    return (sza, Sa)


def _find_indexes(ds, var_name, min_limit, max_limit, use_dask):
    """
    Function to find array indexes where failing limit tests

    Parameters
    ----------
    ds : xarray.Dataset
        Dataset containing data to use in test
    var_name : str
        Variable name to inspect
    min_limit : float or numpy array
        Minimum limit to use for returning indexes
    max_limit : float or numpy array
        Maximum limit to use for returning indexes
    use_dask : boolean
        Option to use Dask operations instead of Numpy

    Returns
    -------
    Tuple containing solar zenith angle array and solar constant scalar

    """
    with warnings.catch_warnings():
        warnings.filterwarnings('ignore', category=RuntimeWarning)
        if use_dask and isinstance(ds[var_name].data, da.Array):
            index_min = da.where(ds[var_name].data < min_limit, True, False).compute()
            index_max = da.where(ds[var_name].data > max_limit, True, False).compute()
        else:
            index_min = np.less(ds[var_name].values, min_limit)
            index_max = np.greater(ds[var_name].values, max_limit)

    return (index_min, index_max)


class QCTests:
    """
    This is a Mixins class used to allow using qcfilter class that is already
    registered to the Xarray dataset. All the methods in this class will be added
    to the qcfilter class. Doing this to make the code spread across more files
    so it is more manageable and readable. Additinal files of tests can be added
    to qcfilter by creating a new class in the new file and adding to qcfilter
    class declaration.

    """

    def bsrn_limits_test(
        self,
        test='Physically Possible',
        gbl_SW_dn_name=None,
        glb_diffuse_SW_dn_name=None,
        direct_normal_SW_dn_name=None,
        direct_SW_dn_name=None,
        glb_SW_up_name=None,
        glb_LW_dn_name=None,
        glb_LW_up_name=None,
        sw_min_limit=None,
        lw_min_dn_limit=None,
        lw_min_up_limit=None,
        lw_max_dn_limit=None,
        lw_max_up_limit=None,
        solar_constant=1366,
        lat_name='lat',
        lon_name='lon',
        use_dask=False,
    ):
        """
        Method to apply BSRN limits test and add results to ancillary quality control variable.
        Need to provide variable name for each measurement for the test to be performed. If no
        limits provided will use default values. All data must be in W/m^2 units.  Test will
        provided exception if required variable name is missing.

        Parameters
        ----------
        test : str or list
            Type of tests to apply. Options include "Physically Possible" or "Extremely Rare"
        gbl_SW_dn_name : str
            Variable name in the Dataset for global shortwave downwelling radiation
            measured by unshaded pyranometer
        glb_diffuse_SW_dn_name : str
            Variable name in the Dataset for global diffuse shortwave downwelling radiation
            measured by shaded pyranometer
        direct_normal_SW_dn_name : str
            Variable name in the Dataset for direct normal shortwave downwelling radiation
        direct_SW_dn_name : str
            Variable name in the Dataset for direct shortwave downwelling radiation
        glb_SW_up_name : str
            Variable name in the Dataset for global shortwave upwelling radiation
        glb_LW_dn_name : str
            Variable name in the Dataset for global longwave downwelling radiation
        glb_LW_up_name : str
            Variable name in the Dataset for global longwave upwelling radiation
        sw_min_limit : int or float
            Lower limit for shortwave radiation test
        lw_min_dn_limit : int or float
            Lower limit for downwelling longwave radiation test measured by a pyrgeometer
        lw_min_up_limit : int or float
            Lower limit for upwelling longwave radiation test measured by a pyrgeometer
        lw_max_dn_limit : int or float
            Upper limit for downwelling longwave radiation test measured by a pyrgeometer
        lw_max_up_limit : int or float
            Upper limit for upwelling longwave radiation test measured by a pyrgeometer
        solar_constant : int or float
            Mean solar constant used in upper limit calculation. Earth sun distance will be
            calculated and applied to this value.
        lat_name : str
            Variable name in the Dataset for latitude
        lon_name : str
            Variable name in the Dataset for longitude
        use_dask : boolean
            Option to use Dask for processing if data is stored in a Dask array

        References
        ----------
        Long, Charles N., and Ellsworth G. Dutton. "BSRN Global Network recommended QC tests, V2. x." (2010).

        Examples
        --------
            .. code-block:: python

                ds = act.io.arm.read_arm_netcdf(act.tests.EXAMPLE_BRS, cleanup_qc=True)
                ds.qcfilter.bsrn_limits_test(
                    gbl_SW_dn_name='down_short_hemisp',
                    glb_diffuse_SW_dn_name='down_short_diffuse_hemisp',
                    direct_normal_SW_dn_name='short_direct_normal',
                    glb_SW_up_name='up_short_hemisp',
                    glb_LW_dn_name='down_long_hemisp_shaded',
                    glb_LW_up_name='up_long_hemisp')
        """

        test_names_org = ["Physically Possible", "Extremely Rare"]
        test = test.lower()
        test_names = [ii.lower() for ii in test_names_org]
        if test not in test_names:
            raise ValueError(
                f"Value of '{test}' in keyword 'test' not recognized. "
                f"Must a single value in options {test_names_org}"
            )

        sza, Sa = _calculate_solar_parameters(self._ds, lat_name, lon_name, solar_constant)

        if test == test_names[0]:
            if sw_min_limit is None:
                sw_min_limit = -4.0
            if lw_min_dn_limit is None:
                lw_min_dn_limit = 40.0
            if lw_min_up_limit is None:
                lw_min_up_limit = 40.0
            if lw_max_dn_limit is None:
                lw_max_dn_limit = 700.0
            if lw_max_up_limit is None:
                lw_max_up_limit = 900.0
        elif test == test_names[1]:
            if sw_min_limit is None:
                sw_min_limit = -2.0
            if lw_min_dn_limit is None:
                lw_min_dn_limit = 60.0
            if lw_min_up_limit is None:
                lw_min_up_limit = 60.0
            if lw_max_dn_limit is None:
                lw_max_dn_limit = 500.0
            if lw_max_up_limit is None:
                lw_max_up_limit = 700.0

        # Global Shortwave downwelling min and max tests
        if gbl_SW_dn_name is not None:
            cos_sza = np.cos(np.radians(sza))
            cos_sza[sza > 90.0] = 0.0
            if test == test_names[0]:
                sw_max_limit = Sa * 1.5 * cos_sza**1.2 + 100.0
            elif test == test_names[1]:
                sw_max_limit = Sa * 1.2 * cos_sza**1.2 + 50.0

            index_min, index_max = _find_indexes(
                self._ds, gbl_SW_dn_name, sw_min_limit, sw_max_limit, use_dask
            )

            self._ds.qcfilter.add_test(
                gbl_SW_dn_name,
                index=index_min,
                test_assessment='Bad',
                test_meaning=f"Value less than BSRN {test.lower()} limit of {sw_min_limit} W/m^2",
            )

            self._ds.qcfilter.add_test(
                gbl_SW_dn_name,
                index=index_max,
                test_assessment='Bad',
                test_meaning=f"Value greater than BSRN {test.lower()} limit",
            )

        # Diffuse Shortwave downwelling min and max tests
        if glb_diffuse_SW_dn_name is not None:
            with warnings.catch_warnings():
                warnings.filterwarnings('ignore', category=RuntimeWarning)
                if test == test_names[0]:
                    sw_max_limit = Sa * 0.95 * np.cos(np.radians(sza)) ** 1.2 + 50.0
                elif test == test_names[1]:
                    sw_max_limit = Sa * 0.75 * np.cos(np.radians(sza)) ** 1.2 + 30.0

            index_min, index_max = _find_indexes(
                self._ds, glb_diffuse_SW_dn_name, sw_min_limit, sw_max_limit, use_dask
            )
            self._ds.qcfilter.add_test(
                glb_diffuse_SW_dn_name,
                index=index_min,
                test_assessment='Bad',
                test_meaning=f"Value less than BSRN {test.lower()} limit of {sw_min_limit} W/m^2",
            )

            self._ds.qcfilter.add_test(
                glb_diffuse_SW_dn_name,
                index=index_max,
                test_assessment='Bad',
                test_meaning=f"Value greater than BSRN {test.lower()} limit",
            )

        # Direct Normal Shortwave downwelling min and max tests
        if direct_normal_SW_dn_name is not None:
            with warnings.catch_warnings():
                warnings.filterwarnings('ignore', category=RuntimeWarning)
                if test == test_names[0]:
                    sw_max_limit = Sa
                elif test == test_names[1]:
                    sw_max_limit = Sa * 0.95 * np.cos(np.radians(sza)) ** 0.2 + 10.0

            index_min, index_max = _find_indexes(
                self._ds, direct_normal_SW_dn_name, sw_min_limit, sw_max_limit, use_dask
            )
            self._ds.qcfilter.add_test(
                direct_normal_SW_dn_name,
                index=index_min,
                test_assessment='Bad',
                test_meaning=f"Value less than BSRN {test.lower()} limit of {sw_min_limit} W/m^2",
            )

            self._ds.qcfilter.add_test(
                direct_normal_SW_dn_name,
                index=index_max,
                test_assessment='Bad',
                test_meaning=f"Value greater than BSRN {test.lower()} limit",
            )

        # Direct Shortwave downwelling min and max tests
        if direct_SW_dn_name is not None:
            with warnings.catch_warnings():
                warnings.filterwarnings('ignore', category=RuntimeWarning)
                if test == test_names[0]:
                    sw_max_limit = Sa * np.cos(np.radians(sza))
                elif test == test_names[1]:
                    sw_max_limit = Sa * 0.95 * np.cos(np.radians(sza)) ** 1.2 + 10

            index_min, index_max = _find_indexes(
                self._ds, direct_SW_dn_name, sw_min_limit, sw_max_limit, use_dask
            )

            self._ds.qcfilter.add_test(
                direct_SW_dn_name,
                index=index_min,
                test_assessment='Bad',
                test_meaning=f"Value less than BSRN {test.lower()} limit of {sw_min_limit} W/m^2",
            )

            self._ds.qcfilter.add_test(
                direct_SW_dn_name,
                index=index_max,
                test_assessment='Bad',
                test_meaning=f"Value greater than BSRN {test.lower()} limit",
            )

        # Shortwave up welling min and max tests
        if glb_SW_up_name is not None:
            with warnings.catch_warnings():
                warnings.filterwarnings('ignore', category=RuntimeWarning)
                if test == test_names[0]:
                    sw_max_limit = Sa * 1.2 * np.cos(np.radians(sza)) ** 1.2 + 50
                elif test == test_names[1]:
                    sw_max_limit = Sa * np.cos(np.radians(sza)) ** 1.2 + 50

            index_min, index_max = _find_indexes(
                self._ds, glb_SW_up_name, sw_min_limit, sw_max_limit, use_dask
            )

            self._ds.qcfilter.add_test(
                glb_SW_up_name,
                index=index_min,
                test_assessment='Bad',
                test_meaning=f"Value less than BSRN {test.lower()} limit of {sw_min_limit} W/m^2",
            )

            self._ds.qcfilter.add_test(
                glb_SW_up_name,
                index=index_max,
                test_assessment='Bad',
                test_meaning=f"Value greater than BSRN {test.lower()} limit",
            )

        # Longwave downwelling min and max tests
        if glb_LW_dn_name is not None:
            index_min, index_max = _find_indexes(
                self._ds, glb_LW_dn_name, lw_min_dn_limit, lw_max_dn_limit, use_dask
            )

            self._ds.qcfilter.add_test(
                glb_LW_dn_name,
                index=index_min,
                test_assessment='Bad',
                test_meaning=f"Value less than BSRN {test.lower()} limit of {lw_min_dn_limit} W/m^2",
            )

            self._ds.qcfilter.add_test(
                glb_LW_dn_name,
                index=index_max,
                test_assessment='Bad',
                test_meaning=f"Value greater than BSRN {test.lower()} limit of {lw_max_dn_limit} W/m^2",
            )

        # Longwave upwelling min and max tests
        if glb_LW_up_name is not None:
            index_min, index_max = _find_indexes(
                self._ds, glb_LW_up_name, lw_min_up_limit, lw_max_up_limit, use_dask
            )

            self._ds.qcfilter.add_test(
                glb_LW_up_name,
                index=index_min,
                test_assessment='Bad',
                test_meaning=f"Value less than BSRN {test.lower()} limit of {lw_min_up_limit} W/m^2",
            )

            self._ds.qcfilter.add_test(
                glb_LW_up_name,
                index=index_max,
                test_assessment='Bad',
                test_meaning=f"Value greater than BSRN {test.lower()} limit of {lw_max_up_limit} W/m^2",
            )

    def bsrn_comparison_tests(
        self,
        test,
        gbl_SW_dn_name=None,
        glb_diffuse_SW_dn_name=None,
        direct_normal_SW_dn_name=None,
        glb_SW_up_name=None,
        glb_LW_dn_name=None,
        glb_LW_up_name=None,
        air_temp_name=None,
        test_assessment='Indeterminate',
        lat_name='lat',
        lon_name='lon',
        LWdn_lt_LWup_component=25.0,
        LWdn_gt_LWup_component=300.0,
        use_dask=False,
    ):
        """
        Method to apply BSRN comparison tests and add results to ancillary quality control variable.
        Need to provided variable name for each measurement for the test to be performed. All radiation
        data must be in W/m^2 units. Test will provided exception if required variable name is missing.

        Parameters
        ----------
        test : str or list
            Type of tests to apply. Options include: 'Global over Sum SW Ratio', 'Diffuse Ratio',
            'SW up', 'LW down to air temp', 'LW up to air temp', 'LW down to LW up', 'Closure'
        gbl_SW_dn_name : str
            Variable name in Dataset for global shortwave downwelling radiation
            measured by unshaded pyranometer
        glb_diffuse_SW_dn_name : str
            Variable name in Dataset for global diffuse shortwave downwelling radiation
            measured by shaded pyranometer
        direct_normal_SW_dn_name : str
            Variable name in Dataset for direct normal shortwave downwelling radiation
        glb_SW_up_name : str
            Variable name in Dataset for global shortwave upwelling radiation
        glb_LW_dn_name : str
            Variable name in Dataset for global longwave downwelling radiation
        glb_LW_up_name : str
            Variable name in Dataset for global longwave upwelling radiation
        air_temp_name : str
            Variable name in Dataset for atmospheric air temperature. Variable used
            in longwave tests.
        test_assessment : str
            Test assessment string value appended to flag_assessments attribute of QC variable.
        lat_name : str
            Variable name in the Dataset for latitude
        lon_name : str
            Variable name in the Dataset for longitude
        LWdn_lt_LWup_component : int or float
            Value used in longwave down less than longwave up test.
        LWdn_gt_LWup_component : int or float
            Value used in longwave down greater than longwave up test.
        use_dask : bool
            Option to use Dask for processing if data is stored in a Dask array

        References
        ----------
        Long, Charles N., and Ellsworth G. Dutton. "BSRN Global Network recommended QC tests, V2. x." (2010).
        Long, Charles N., and Y. Shi.  “An Automated Quality Assessment and Control Algorithm for Surface
        Radiation Measurements.” (2008) https://doi.org/10.2174/1874282300802010023

        Examples
        --------

        An example on how to add BSRN comparison tests:

        .. code-block:: python

            ds = act.io.arm.read_arm_netcdf(act.tests.EXAMPLE_BRS, cleanup_qc=True)
            ds.qcfilter.bsrn_comparison_tests(
                gbl_SW_dn_name='down_short_hemisp',
                glb_diffuse_SW_dn_name='down_short_diffuse_hemisp',
                direct_normal_SW_dn_name='short_direct_normal',
                glb_SW_up_name='up_short_hemisp',
                glb_LW_dn_name='down_long_hemisp_shaded',
                glb_LW_up_name='up_long_hemisp',
                use_dask=True
            )

        """
        if isinstance(test, str):
            test = [test]

        test = [ii.lower() for ii in test]

        test_options = [
            'Global over Sum SW Ratio',
            'Diffuse Ratio',
            'SW up',
            'LW down to air temp',
            'LW up to air temp',
            'LW down to LW up',
            'Closure',
        ]

        solar_constant = 1360.8
        sza, Sa = _calculate_solar_parameters(self._ds, lat_name, lon_name, solar_constant)

        # Ratio of Global over Sum SW
        if test_options[0].lower() in test:
            if (
                gbl_SW_dn_name is None
                or glb_diffuse_SW_dn_name is None
                or direct_normal_SW_dn_name is None
            ):
                raise ValueError(
                    'Must set keywords gbl_SW_dn_name, glb_diffuse_SW_dn_name, '
                    f'direct_normal_SW_dn_name for {test_options[0]} test.'
                )

            with warnings.catch_warnings():
                warnings.filterwarnings('ignore', category=RuntimeWarning)
                if use_dask and isinstance(self._ds[glb_diffuse_SW_dn_name].data, da.Array):
                    sum_sw_down = self._ds[glb_diffuse_SW_dn_name].data + self._ds[
                        direct_normal_SW_dn_name
                    ].data * np.cos(np.radians(sza))
                    sum_sw_down[sum_sw_down < 50] = np.nan
                    ratio = self._ds[gbl_SW_dn_name].data / sum_sw_down
                    index_a = sza < 75
                    index_1 = da.where((ratio > 1.08) & index_a, True, False)
                    index_2 = da.where((ratio < 0.92) & index_a, True, False)
                    index_b = (sza >= 75) & (sza < 93)
                    index_3 = da.where((ratio > 1.15) & index_b & index_b, True, False)
                    index_4 = da.where((ratio < 0.85) & index_b, True, False)
                    index = (index_1 | index_2 | index_3 | index_4).compute()
                else:
                    sum_sw_down = self._ds[glb_diffuse_SW_dn_name].values + self._ds[
                        direct_normal_SW_dn_name
                    ].values * np.cos(np.radians(sza))
                    sum_sw_down[sum_sw_down < 50] = np.nan
                    ratio = self._ds[gbl_SW_dn_name].values / sum_sw_down
                    index_a = sza < 75
                    index_1 = (ratio > 1.08) & index_a
                    index_2 = (ratio < 0.92) & index_a
                    index_b = (sza >= 75) & (sza < 93)
                    index_3 = (ratio > 1.15) & index_b
                    index_4 = (ratio < 0.85) & index_b
                    index = index_1 | index_2 | index_3 | index_4

            test_meaning = "Ratio of Global over Sum shortwave larger than expected"
            self._ds.qcfilter.add_test(
                gbl_SW_dn_name,
                index=index,
                test_assessment=test_assessment,
                test_meaning=test_meaning,
            )
            self._ds.qcfilter.add_test(
                glb_diffuse_SW_dn_name,
                index=index,
                test_assessment=test_assessment,
                test_meaning=test_meaning,
            )
            self._ds.qcfilter.add_test(
                direct_normal_SW_dn_name,
                index=index,
                test_assessment=test_assessment,
                test_meaning=test_meaning,
            )

        # Diffuse Ratio
        if test_options[1].lower() in test:
            if gbl_SW_dn_name is None or glb_diffuse_SW_dn_name is None:
                raise ValueError(
                    'Must set keywords gbl_SW_dn_name, glb_diffuse_SW_dn_name '
                    f'for {test_options[1]} test.'
                )

            with warnings.catch_warnings():
                warnings.filterwarnings('ignore', category=RuntimeWarning)
                if use_dask and isinstance(self._ds[glb_diffuse_SW_dn_name].data, da.Array):
                    ratio = self._ds[glb_diffuse_SW_dn_name].data / self._ds[gbl_SW_dn_name].data
                    ratio[self._ds[gbl_SW_dn_name].data < 50] = np.nan
                    index_a = sza < 75
                    index_1 = da.where((ratio >= 1.05) & index_a, True, False)
                    index_b = (sza >= 75) & (sza < 93)
                    index_2 = da.where((ratio >= 1.10) & index_b, True, False)
                    index = (index_1 | index_2).compute()
                else:
                    ratio = (
                        self._ds[glb_diffuse_SW_dn_name].values / self._ds[gbl_SW_dn_name].values
                    )
                    ratio[self._ds[gbl_SW_dn_name].values < 50] = np.nan
                    index_a = sza < 75
                    index_1 = (ratio >= 1.05) & index_a
                    index_b = (sza >= 75) & (sza < 93)
                    index_2 = (ratio >= 1.10) & index_b
                    index = index_1 | index_2

            test_meaning = "Ratio of Diffuse Shortwave over Global Shortwave larger than expected"
            self._ds.qcfilter.add_test(
                gbl_SW_dn_name,
                index=index,
                test_assessment=test_assessment,
                test_meaning=test_meaning,
            )
            self._ds.qcfilter.add_test(
                glb_diffuse_SW_dn_name,
                index=index,
                test_assessment=test_assessment,
                test_meaning=test_meaning,
            )

        # Shortwave up comparison
        if test_options[2].lower() in test:
            if (
                glb_SW_up_name is None
                or glb_diffuse_SW_dn_name is None
                or direct_normal_SW_dn_name is None
            ):
                raise ValueError(
                    'Must set keywords glb_SW_up_name, glb_diffuse_SW_dn_name, '
                    f'direct_normal_SW_dn_name for {test_options[2]} test.'
                )

            with warnings.catch_warnings():
                warnings.filterwarnings('ignore', category=RuntimeWarning)
                if use_dask and isinstance(self._ds[glb_diffuse_SW_dn_name].data, da.Array):
                    sum_sw_down = self._ds[glb_diffuse_SW_dn_name].data + self._ds[
                        direct_normal_SW_dn_name
                    ].data * np.cos(np.radians(sza))

                    sum_sw_down[sum_sw_down < 50] = np.nan
                    index = da.where(
                        self._ds[glb_SW_up_name].data > sum_sw_down, True, False
                    ).compute()
                else:
                    sum_sw_down = self._ds[glb_diffuse_SW_dn_name].values + self._ds[
                        direct_normal_SW_dn_name
                    ].values * np.cos(np.radians(sza))
                    sum_sw_down[sum_sw_down < 50] = np.nan
                    index = self._ds[glb_SW_up_name].values > sum_sw_down

            test_meaning = "Ratio of Shortwave Upwelling greater than Shortwave Sum"
            self._ds.qcfilter.add_test(
                glb_SW_up_name,
                index=index,
                test_assessment=test_assessment,
                test_meaning=test_meaning,
            )
            self._ds.qcfilter.add_test(
                glb_diffuse_SW_dn_name,
                index=index,
                test_assessment=test_assessment,
                test_meaning=test_meaning,
            )
            self._ds.qcfilter.add_test(
                direct_normal_SW_dn_name,
                index=index,
                test_assessment=test_assessment,
                test_meaning=test_meaning,
            )

        # Longwave down to air temperature comparison
        if test_options[3].lower() in test:
            if glb_LW_dn_name is None or air_temp_name is None:
                raise ValueError(
                    'Must set keywords glb_LW_dn_name, air_temp_name '
                    f' for {test_options[3]} test.'
                )

            air_temp = convert_units(
                self._ds[air_temp_name].values, self._ds[air_temp_name].attrs['units'], 'degK'
            )
            if use_dask and isinstance(self._ds[glb_LW_dn_name].data, da.Array):
                air_temp = da.array(air_temp)
                conversion = da.array(Stefan_Boltzmann * air_temp**4)
                index_1 = (0.4 * conversion) > self._ds[glb_LW_dn_name].data
                index_2 = (conversion + 25.0) < self._ds[glb_LW_dn_name].data
                index = (index_1 | index_2).compute()
            else:
                conversion = Stefan_Boltzmann * air_temp**4
                index_1 = (0.4 * conversion) > self._ds[glb_LW_dn_name].values
                index_2 = (conversion + 25.0) < self._ds[glb_LW_dn_name].values
                index = index_1 | index_2

            test_meaning = (
                "Longwave downwelling comparison to air temperature out side of expected range"
            )
            self._ds.qcfilter.add_test(
                glb_LW_dn_name,
                index=index,
                test_assessment=test_assessment,
                test_meaning=test_meaning,
            )

        # Longwave up to air temperature comparison
        if test_options[4].lower() in test:
            if glb_LW_up_name is None or air_temp_name is None:
                raise ValueError(
                    'Must set keywords glb_LW_up_name, air_temp_name '
                    f'for {test_options[3]} test.'
                )

            air_temp = convert_units(
                self._ds[air_temp_name].values, self._ds[air_temp_name].attrs['units'], 'degK'
            )
            if use_dask and isinstance(self._ds[glb_LW_up_name].data, da.Array):
                air_temp = da.array(air_temp)
                index_1 = (Stefan_Boltzmann * (air_temp - 15) ** 4) > self._ds[glb_LW_up_name].data
                index_2 = (Stefan_Boltzmann * (air_temp + 25) ** 4) < self._ds[glb_LW_up_name].data
                index = (index_1 | index_2).compute()
            else:
                index_1 = (Stefan_Boltzmann * (air_temp - 15) ** 4) > self._ds[
                    glb_LW_up_name
                ].values
                index_2 = (Stefan_Boltzmann * (air_temp + 25) ** 4) < self._ds[
                    glb_LW_up_name
                ].values
                index = index_1 | index_2

            test_meaning = (
                "Longwave upwelling comparison to air temperature out side of expected range"
            )
            self._ds.qcfilter.add_test(
                glb_LW_up_name,
                index=index,
                test_assessment=test_assessment,
                test_meaning=test_meaning,
            )

        # Lonwave down to longwave up comparison
        if test_options[5].lower() in test:
            if glb_LW_dn_name is None or glb_LW_up_name is None:
                raise ValueError(
                    'Must set keywords glb_LW_dn_name, glb_LW_up_name '
                    f'for {test_options[3]} test.'
                )

            if use_dask and isinstance(self._ds[glb_LW_dn_name].data, da.Array):
                index_1 = da.where(
                    self._ds[glb_LW_dn_name].data
                    > (self._ds[glb_LW_up_name].data + LWdn_lt_LWup_component),
                    True,
                    False,
                )
                index_2 = da.where(
                    self._ds[glb_LW_dn_name].data
                    < (self._ds[glb_LW_up_name].data - LWdn_gt_LWup_component),
                    True,
                    False,
                )
                index = (index_1 | index_2).compute()
            else:
                index_1 = self._ds[glb_LW_dn_name].values > (
                    self._ds[glb_LW_up_name].values + LWdn_lt_LWup_component
                )
                index_2 = self._ds[glb_LW_dn_name].values < (
                    self._ds[glb_LW_up_name].values - LWdn_gt_LWup_component
                )
                index = index_1 | index_2

            test_meaning = (
                "Lonwave downwelling compared to longwave upwelling outside of expected range"
            )
            self._ds.qcfilter.add_test(
                glb_LW_dn_name,
                index=index,
                test_assessment=test_assessment,
                test_meaning=test_meaning,
            )
            self._ds.qcfilter.add_test(
                glb_LW_up_name,
                index=index,
                test_assessment=test_assessment,
                test_meaning=test_meaning,
            )

        # Closure test
        if test_options[6].lower() in test:
            if (
                direct_normal_SW_dn_name is None
                or glb_diffuse_SW_dn_name is None
                or gbl_SW_dn_name is None
            ):
                raise ValueError(
                    'Must set keywords direct_normal_SW_dn_name, glb_diffuse_SW_dn_name, '
                    f'gbl_SW_dn_name for {test_options[3]} test.'
                )

            if use_dask and isinstance(self._ds[gbl_SW_dn_name].data, da.Array):
                sza = da.array(sza)
                component_sum = self._ds[gbl_SW_dn_name].values + (
                    self._ds[direct_normal_SW_dn_name].values * np.cos(np.radians(sza))
                )

                index_lsz = da.where((component_sum > 50.0) & (sza <= 75.0), True, False)
                index_hsz = da.where(
                    (component_sum > 50.0) & (sza > 75.0) & (sza < 93.0), True, False
                )

                value = (self._ds[glb_diffuse_SW_dn_name].values / component_sum) - 1.0

                index_1 = da.where((value >= 0.08) & index_lsz, True, False)
                index_2 = da.where((value >= 0.15) & index_hsz, True, False)

                index = (index_1 | index_2).compute()

            else:
                component_sum = self._ds[gbl_SW_dn_name].values + (
                    self._ds[direct_normal_SW_dn_name].values * np.cos(np.radians(sza))
                )

                index_lsz = (component_sum > 50.0) & (sza <= 75.0)
                index_hsz = (component_sum > 50.0) & (sza > 75.0) & (sza < 93.0)

                value = (self._ds[glb_diffuse_SW_dn_name].values / component_sum) - 1.0

                index_1 = (value >= 0.08) & index_lsz
                index_2 = (value >= 0.15) & index_hsz

                index = index_1 | index_2

            test_meaning = "Closure test indicating value outside of expected range"
            self._ds.qcfilter.add_test(
                direct_normal_SW_dn_name,
                index=index,
                test_assessment=test_assessment,
                test_meaning=test_meaning,
            )
            self._ds.qcfilter.add_test(
                glb_diffuse_SW_dn_name,
                index=index,
                test_assessment=test_assessment,
                test_meaning=test_meaning,
            )
            self._ds.qcfilter.add_test(
                gbl_SW_dn_name,
                index=index,
                test_assessment=test_assessment,
                test_meaning=test_meaning,
            )

    def normalized_rradiance_test(
        self,
        test,
        dni=None,
        dhi=None,
        ghi=None,
        test_assessment='Indeterminate',
        lat_name='lat',
        lon_name='lon',
        alt_name='alt',
        upper_total_transmittance_limit=1.6,
        upper_diffuse_transmittance_limit=1.0,
        use_dask=False,
    ):
        """
        Method to apply Normalized Irradiance Tests tests and add results to ancillary quality control variable.
        Need to provided variable name for each measurement for the test to be performed. All radiation
        data must be in W/m^2 units. All results from tests are set in ancillary quality control varibles
        for dni and dhi.

        test : list or str
            Type of tests to apply. Options include: 'Clearness index', 'Upper total transmittance',
            'Upper direct transmittance', 'Upper diffuse transmittance'
        dni : str
            Variable name in Dataset for direct normal irradiance
        dhi : str
            Variable name in Dataset for diffuse hemipheric irradiance
        ghi : str
            Variable name in Dataset for global hemispheric irradiance. Used in
            'Upper direct transmittance' test only.
        test_assessment : str
            Test assessment string value appended to flag_assessments attribute of QC variable.
        lat_name : str
            Variable name in the Dataset for latitude
        lon_name : str
            Variable name in the Dataset for longitude
        alt_name : str
            Variable name in the Dataset for altitude. Used in 'Upper direct transmittance' test.
        upper_total_transmittance_limit : float
            Limit value used in 'Upper total transmittance' test. Values larger than this will be flagged.
        upper_diffuse_transmittance_limit : float
            Limit value used in 'Upper diffuse transmittance' test. Values larger than this will be flagged.
        use_dask : boolean
            Option to use Dask for processing if data is stored in a Dask array


        References
        ----------
        Sengupta, Manajit, Habte, Aron, Wilbert, Stefan , Christian Gueymard, Jan Remund, Elke Lorenz,
            Wilfried van Sark, and Adam R. Jensen "Best Practices Handbook for the Collection and Use of Solar
            Resource Data for Solar Energy Applications: Fourth Edition" (2024).

        Examples
        --------
            .. code-block:: python

                ds = act.io.arm.read_arm_netcdf(act.tests.EXAMPLE_BRS, cleanup_qc=True)
                ds.qcfilter.normalized_rradiance_test(
                    ['Clearness index', 'Upper total transmittance',
                     'Upper direct transmittance', 'Upper diffuse transmittance'],
                    dhi='down_short_diffuse_hemisp',
                    ghi='down_short_hemisp',
                    dni='short_direct_normal'
                )
        """

        if isinstance(test, str):
            test = [test]

        test = [ii.lower() for ii in test]

        solar_constant = 1366
        sza, Sa = _calculate_solar_parameters(self._ds, lat_name, lon_name, solar_constant)

        test_options = [
            'Clearness index',
            'Upper total transmittance',
            'Upper direct transmittance',
            'Upper diffuse transmittance',
        ]

        if test_options[0].lower() in test:
            if dni is None:
                raise RuntimeError(
                    "Need to set 'dni' keyword to perform 'Clearness index' test in normalized_rradiance_test()."
                )
            if dhi is None:
                raise RuntimeError(
                    "Need to set 'dhi' keyword to perform 'Clearness index' test in normalized_rradiance_test()."
                )

            if use_dask and isinstance(self._ds[dni].data, da.Array):
                sza = da.array(sza)
                Sa = da.array(Sa)
                K_direct = self._ds[dni].values / Sa
                K_diffuse = self._ds[dhi].values / (Sa * np.cos(np.radians(sza)))
                K_total = K_direct + K_diffuse
                index = (K_direct > K_total).compute()

            else:
                K_direct = self._ds[dni].values / Sa
                K_diffuse = self._ds[dhi].values / (Sa * np.cos(np.radians(sza)))
                K_total = K_direct + K_diffuse
                index = K_direct > K_total

            test_meaning = "Normalized direct normal irradiance greater than total transmittance."
            self._ds.qcfilter.add_test(
                dni,
                index=index,
                test_assessment=test_assessment,
                test_meaning=test_meaning,
            )
            self._ds.qcfilter.add_test(
                dhi,
                index=index,
                test_assessment=test_assessment,
                test_meaning=test_meaning,
            )

        if test_options[1].lower() in test:
            if dni is None:
                raise RuntimeError(
                    "Need to set 'dni' keyword to perform 'Upper total transmittance' test in normalized_rradiance_test()."
                )
            if dhi is None:
                raise RuntimeError(
                    "Need to set 'dhi' keyword to perform 'Upper total transmittance' test in normalized_rradiance_test()."
                )

            if use_dask and isinstance(self._ds[dni].data, da.Array):
                sza = da.array(sza)
                Sa = da.array(Sa)
                K_direct = self._ds[dni].values / Sa
                K_diffuse = self._ds[dhi].values / (Sa * np.cos(np.radians(sza)))
                K_total = K_direct + K_diffuse
                index = (K_total > upper_total_transmittance_limit) & (sza > 0) & (sza < 90)
                index = index.compute()

            else:
                K_direct = self._ds[dni].values / Sa
                K_diffuse = self._ds[dhi].values / (Sa * np.cos(np.radians(sza)))
                K_total = K_direct + K_diffuse
                index = (K_total > upper_total_transmittance_limit) & (sza > 0) & (sza < 90)

            test_meaning = f"Total transmittance greater than {upper_total_transmittance_limit}"
            self._ds.qcfilter.add_test(
                dni,
                index=index,
                test_assessment=test_assessment,
                test_meaning=test_meaning,
            )
            self._ds.qcfilter.add_test(
                dhi,
                index=index,
                test_assessment=test_assessment,
                test_meaning=test_meaning,
            )

        if test_options[2].lower() in test:
            if dni is None:
                raise RuntimeError(
                    "Need to set 'dni' keyword to perform 'Upper direct transmittance' test in normalized_rradiance_test()."
                )
            if ghi is None:
                raise RuntimeError(
                    "Need to set 'ghi' keyword to perform 'Upper direct transmittance' test in normalized_rradiance_test()."
                )

            if use_dask and isinstance(self._ds[dni].data, da.Array):
                Sa = da.array(Sa)
                K_direct = self._ds[dni].values / Sa
                alt = da.array(
                    convert_units(self._ds[alt_name].values, self._ds[alt_name].attrs['units'], 'm')
                )
                index1 = (K_direct > 0) & (self._ds[ghi].values > 50.0)
                index2 = K_direct > ((1100 + 0.03 * alt) / Sa)
                index = (index1 & index2).compute()

            else:
                K_direct = self._ds[dni].values / Sa
                alt = convert_units(
                    self._ds[alt_name].values, self._ds[alt_name].attrs['units'], 'm'
                )
                index1 = (K_direct > 0) & (self._ds[ghi].values > 50.0)
                index2 = K_direct > ((1100 + 0.03 * alt) / Sa)
                index = index1 & index2

            test_meaning = "Direct transmittance greater than upper direct transmittance limit"
            self._ds.qcfilter.add_test(
                dni,
                index=index,
                test_assessment=test_assessment,
                test_meaning=test_meaning,
            )

        if test_options[3].lower() in test:
            if dhi is None:
                raise RuntimeError(
                    "Need to set 'dhi' keyword to perform 'Upper diffuse transmittance' test in normalized_rradiance_test()."
                )

            if use_dask and isinstance(self._ds[dhi].data, da.Array):
                sza = da.array(sza)
                Sa = da.array(Sa)
                K_diffuse = self._ds[dhi].values / (Sa * np.cos(np.radians(sza)))
                index = (sza > 0) & (sza < 90) & (K_diffuse > upper_diffuse_transmittance_limit)
                index = index.compute()

            else:
                K_diffuse = self._ds[dhi].values / (Sa * np.cos(np.radians(sza)))
                index = (sza > 0) & (sza < 90) & (K_diffuse > upper_diffuse_transmittance_limit)

            test_meaning = f"Diffuse transmittance greater than {upper_diffuse_transmittance_limit}"
            self._ds.qcfilter.add_test(
                dhi,
                index=index,
                test_assessment=test_assessment,
                test_meaning=test_meaning,
            )



import matplotlib
import numpy as np
import pytest

import act
from act.plotting import WindRoseDisplay
from act.tests import sample_files

matplotlib.use('Agg')


@pytest.mark.mpl_image_compare(tolerance=10)
def test_wind_rose():
    sonde_ds = act.io.arm.read_arm_netcdf(sample_files.EXAMPLE_TWP_SONDE_WILDCARD)

    WindDisplay = WindRoseDisplay(sonde_ds, figsize=(10, 10))
    WindDisplay.plot(
        'deg',
        'wspd',
        spd_bins=np.linspace(0, 20, 10),
        num_dirs=30,
        tick_interval=2,
        cmap='viridis',
    )
    WindDisplay.set_thetarng(trng=(0.0, 360.0))
    WindDisplay.set_rrng((0.0, 14))

    sonde_ds.close()

    try:
        return WindDisplay.fig
    finally:
        matplotlib.pyplot.close(WindDisplay.fig)


@pytest.mark.mpl_image_compare(tolerance=10)
def test_plot_datarose():
    files = sample_files.EXAMPLE_MET_WILDCARD
    ds = act.io.arm.read_arm_netcdf(files)
    display = act.plotting.WindRoseDisplay(ds, subplot_shape=(2, 3), figsize=(16, 10))
    display.plot_data(
        'wdir_vec_mean',
        'wspd_vec_mean',
        'temp_mean',
        num_dirs=12,
        plot_type='line',
        subplot_index=(0, 0),
    )
    display.plot_data(
        'wdir_vec_mean',
        'wspd_vec_mean',
        'temp_mean',
        num_dirs=12,
        plot_type='line',
        subplot_index=(0, 1),
        line_plot_calc='median',
    )
    display.plot_data(
        'wdir_vec_mean',
        'wspd_vec_mean',
        'temp_mean',
        num_dirs=12,
        plot_type='line',
        subplot_index=(0, 2),
        line_plot_calc='stdev',
    )
    display.plot_data(
        'wdir_vec_mean',
        'wspd_vec_mean',
        'temp_mean',
        num_dirs=12,
        plot_type='contour',
        subplot_index=(1, 0),
    )
    display.plot_data(
        'wdir_vec_mean',
        'wspd_vec_mean',
        'temp_mean',
        num_dirs=12,
        plot_type='contour',
        contour_type='mean',
        num_data_bins=10,
        clevels=21,
        cmap='rainbow',
        vmin=-5,
        vmax=20,
        subplot_index=(1, 1),
    )
    display.plot_data(
        'wdir_vec_mean',
        'wspd_vec_mean',
        'temp_mean',
        num_dirs=12,
        plot_type='boxplot',
        subplot_index=(1, 2),
    )

    display2 = act.plotting.WindRoseDisplay(
        {'ds1': ds, 'ds2': ds}, subplot_shape=(2, 3), figsize=(16, 10)
    )
    with pytest.raises(ValueError):
        display2.plot_data(
            'wdir_vec_mean',
            'wspd_vec_mean',
            'temp_mean',
            dsname='ds1',
            num_dirs=12,
            plot_type='line',
            line_plot_calc='T',
            subplot_index=(0, 0),
        )
    with pytest.raises(ValueError):
        display2.plot_data(
            'wdir_vec_mean',
            'wspd_vec_mean',
            'temp_mean',
            num_dirs=12,
            plot_type='line',
            subplot_index=(0, 0),
        )
    with pytest.raises(ValueError):
        display2.plot_data(
            'wdir_vec_mean',
            'wspd_vec_mean',
            'temp_mean',
            num_dirs=12,
            dsname='ds1',
            plot_type='groovy',
            subplot_index=(0, 0),
        )

    return display.fig


@pytest.mark.mpl_image_compare(tolerance=10)
def test_missing_axes_fig():
    sonde_ds = act.io.arm.read_arm_netcdf(sample_files.EXAMPLE_TWP_SONDE_WILDCARD)

    WindDisplay = WindRoseDisplay(sonde_ds, figsize=(10, 10))
    WindDisplay.fig = None
    WindDisplay.axes = None
    WindDisplay.plot(
        'deg',
        'wspd',
        spd_bins=None,
        num_dirs=30,
        tick_interval=2,
        cmap='viridis',
    )
    try:
        return WindDisplay.fig
    finally:
        matplotlib.pyplot.close(WindDisplay.fig)


@pytest.mark.mpl_image_compare(tolerance=10)
def test_groupby_plot():
    ds = act.io.arm.read_arm_netcdf(act.tests.EXAMPLE_MET_WILDCARD)

    # Create Plot Display
    display = WindRoseDisplay(ds, figsize=(15, 15), subplot_shape=(3, 3))
    groupby = display.group_by('day')
    with pytest.warns(RuntimeWarning):
        groupby.plot_group(
            'plot_data',
            None,
            dir_field='wdir_vec_mean',
            spd_field='wspd_vec_mean',
            data_field='temp_mean',
            num_dirs=12,
            plot_type='line',
        )

    # Set theta tick markers for each axis inside display to be inside the polar axes
    for i in range(3):
        for j in range(3):
            display.axes[i, j].tick_params(pad=-20)
    ds.close()
    return display.fig


def test_windrose_errors():
    sonde_ds = act.io.arm.read_arm_netcdf(sample_files.EXAMPLE_TWP_SONDE_WILDCARD)
    WindDisplay = WindRoseDisplay(sonde_ds, figsize=(10, 10))
    WindDisplay.axes = None
    pytest.raises(RuntimeError, WindDisplay.set_thetarng)
    pytest.raises(RuntimeError, WindDisplay.set_rrng, (50, 80))



"""
Plotting flagged data gaps with a step change test
--------------------------------------------------

This is an example for how to use the step change detection test.
The test uses the cumulative sum control chart to detect when
a sudden shift in values occurs. It has an option to insert
NaN value when there is a data gap to not have those periods
returned as a data shift. This example produces two plots,
one with the data gap flagged and one without.

Author: Ken Kehoe

"""

import numpy as np
from arm_test_data import DATASETS
from matplotlib import pyplot as plt

from act.io.arm import read_arm_netcdf

# Get example data from ARM Test Data repository
EXAMPLE_MET = DATASETS.fetch('sgpmetE13.b1.20190101.000000.cdf')
variable = 'temp_mean'
ds = read_arm_netcdf(EXAMPLE_MET, keep_variables=variable)

# Add shifts in the data
data = ds[variable].values
data[600:] += 2
data[1000:] -= 2
ds[variable].values = data

# Remove data from the Dataset to simulate instrument being off-line
ds = ds.where((ds["time.hour"] < 3) | (ds["time.hour"] > 5), drop=True)

# Add step change test
ds.qcfilter.add_step_change_test(variable)

# Add step change test but insert NaN values during period of missing data
# so it does not trip the test.
ds.qcfilter.add_step_change_test(variable, add_nan=True)

# Make plot with results from the step change test for when the missing data
# is included and a second plot without including the missing data gap.
title = 'Step change detection'
for ii in range(1, 3):
    plt.figure(figsize=(10, 6))
    plt.plot(ds['time'].values, ds[variable].values, label='Data')
    plt.xlabel('Time')
    plt.ylabel(f"{ds[variable].attrs['long_name']} ({ds[variable].attrs['units']})")
    plt.title(title)
    plt.grid(lw=2, ls=':')

    label = 'Step change'
    index = ds.qcfilter.get_qc_test_mask(var_name=variable, test_number=ii)
    for jj in np.where(index)[0]:
        plt.axvline(x=ds['time'].values[jj], color='orange', linestyle='--', label=label)
        label = None

    title += ' with NaN added in data gaps'

    plt.legend()
    plt.show()
